# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from keystoneauth1 import adapter

from openstack.identity.v3 import group
from openstack.identity.v3 import user
from openstack.tests.unit import base


IDENTIFIER = 'IDENTIFIER'
EXAMPLE = {
    'description': '1',
    'domain_id': '2',
    'id': IDENTIFIER,
    'name': '4',
}


class TestGroup(base.TestCase):

    def setUp(self):
        super(TestGroup, self).setUp()
        self.sess = mock.Mock(spec=adapter.Adapter)
        self.sess.default_microversion = 1
        self.sess._get_connection = mock.Mock(return_value=self.cloud)
        self.good_resp = mock.Mock()
        self.good_resp.body = None
        self.good_resp.json = mock.Mock(return_value=self.good_resp.body)
        self.good_resp.status_code = 204

    def test_basic(self):
        sot = group.Group()
        self.assertEqual('group', sot.resource_key)
        self.assertEqual('groups', sot.resources_key)
        self.assertEqual('/groups', sot.base_path)
        self.assertTrue(sot.allow_create)
        self.assertTrue(sot.allow_fetch)
        self.assertTrue(sot.allow_commit)
        self.assertTrue(sot.allow_delete)
        self.assertTrue(sot.allow_list)
        self.assertEqual('PATCH', sot.commit_method)

        self.assertDictEqual(
            {
                'domain_id': 'domain_id',
                'name': 'name',
                'limit': 'limit',
                'marker': 'marker',
            },
            sot._query_mapping._mapping)

    def test_make_it(self):
        sot = group.Group(**EXAMPLE)
        self.assertEqual(EXAMPLE['description'], sot.description)
        self.assertEqual(EXAMPLE['domain_id'], sot.domain_id)
        self.assertEqual(EXAMPLE['id'], sot.id)
        self.assertEqual(EXAMPLE['name'], sot.name)

    def test_add_user(self):
        sot = group.Group(**EXAMPLE)
        resp = self.good_resp
        self.sess.put = mock.Mock(return_value=resp)

        sot.add_user(
            self.sess, user.User(id='1'))

        self.sess.put.assert_called_with(
            'groups/IDENTIFIER/users/1')

    def test_remove_user(self):
        sot = group.Group(**EXAMPLE)
        resp = self.good_resp
        self.sess.delete = mock.Mock(return_value=resp)

        sot.remove_user(
            self.sess, user.User(id='1'))

        self.sess.delete.assert_called_with(
            'groups/IDENTIFIER/users/1')

    def test_check_user(self):
        sot = group.Group(**EXAMPLE)
        resp = self.good_resp
        self.sess.head = mock.Mock(return_value=resp)

        self.assertTrue(
            sot.check_user(
                self.sess,
                user.User(id='1')))

        self.sess.head.assert_called_with(
            'groups/IDENTIFIER/users/1')



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack import resource


class RoleAssignment(resource.Resource):
    resource_key = 'role_assignment'
    resources_key = 'role_assignments'
    base_path = '/role_assignments'

    # capabilities
    allow_list = True

    _query_mapping = resource.QueryParameters(
        'group_id', 'role_id', 'scope_domain_id', 'scope_project_id',
        'user_id', 'effective', 'include_names', 'include_subtree',
        role_id='role.id', user_id='user.id', group_id='group.id',
        scope_project_id='scope.project.id', scope_domain_id='scope.domain.id',
        scope_system='scope.system',
    )

    # Properties
    #: The links for the service resource.
    links = resource.Body('links')
    #: The role (dictionary contains only id) *Type: dict*
    role = resource.Body('role', type=dict)
    #: The scope (either domain or group dictionary contains id) *Type: dict*
    scope = resource.Body('scope', type=dict)
    #: The user (dictionary contains only id) *Type: dict*
    user = resource.Body('user', type=dict)
    #: The group (dictionary contains only id) *Type: dict*
    group = resource.Body('group', type=dict)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack import exceptions
from openstack import resource
from openstack import utils


class AddressGroup(resource.Resource):
    """Address group extension."""
    resource_key = 'address_group'
    resources_key = 'address_groups'
    base_path = '/address-groups'

    # capabilities
    allow_create = True
    allow_fetch = True
    allow_commit = True
    allow_delete = True
    allow_list = True

    _allow_unknown_attrs_in_body = True

    _query_mapping = resource.QueryParameters(
        "sort_key", "sort_dir",
        'name', 'description',
        'project_id'
    )

    # Properties
    #: The ID of the address group.
    id = resource.Body('id')
    #: The address group name.
    name = resource.Body('name')
    #: The address group name.
    description = resource.Body('description')
    #: The ID of the project that owns the address group.
    project_id = resource.Body('project_id', alias='tenant_id')
    #: Tenant_id (deprecated attribute).
    tenant_id = resource.Body('tenant_id', deprecated=True)
    #: The IP addresses of the address group.
    addresses = resource.Body('addresses', type=list)

    def _put(self, session, url, body):
        resp = session.put(url, json=body)
        exceptions.raise_from_response(resp)
        return resp

    def add_addresses(self, session, addresses):
        """Add addresses into the address group.

        :param session: The session to communicate through.
        :type session: :class:`~keystoneauth1.adapter.Adapter`
        :param list addresses: The list of address strings.

        :returns: The response as a AddressGroup object with updated addresses

        :raises: :class:`~openstack.exceptions.SDKException` on error.
        """
        url = utils.urljoin(self.base_path, self.id, 'add_addresses')
        resp = self._put(session, url, {'addresses': addresses})
        self._translate_response(resp)
        return self

    def remove_addresses(self, session, addresses):
        """Remove addresses from the address group.

        :param session: The session to communicate through.
        :type session: :class:`~keystoneauth1.adapter.Adapter`
        :param list addresses: The list of address strings.

        :returns: The response as a AddressGroup object with updated addresses

        :raises: :class:`~openstack.exceptions.SDKException` on error.
        """
        url = utils.urljoin(self.base_path, self.id, 'remove_addresses')
        resp = self._put(session, url, {'addresses': addresses})
        self._translate_response(resp)
        return self



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.tests.functional import base


class BaseBaremetalTest(base.BaseFunctionalTest):

    min_microversion = None
    node_id = None

    def setUp(self):
        super(BaseBaremetalTest, self).setUp()
        self.require_service('baremetal',
                             min_microversion=self.min_microversion)

    def create_allocation(self, **kwargs):
        allocation = self.conn.baremetal.create_allocation(**kwargs)
        self.addCleanup(
            lambda: self.conn.baremetal.delete_allocation(allocation.id,
                                                          ignore_missing=True))
        return allocation

    def create_chassis(self, **kwargs):
        chassis = self.conn.baremetal.create_chassis(**kwargs)
        self.addCleanup(
            lambda: self.conn.baremetal.delete_chassis(chassis.id,
                                                       ignore_missing=True))
        return chassis

    def create_node(self, driver='fake-hardware', **kwargs):
        node = self.conn.baremetal.create_node(driver=driver, **kwargs)
        self.node_id = node.id
        self.addCleanup(
            lambda: self.conn.baremetal.delete_node(self.node_id,
                                                    ignore_missing=True))
        self.assertIsNotNone(self.node_id)
        return node

    def create_port(self, node_id=None, **kwargs):
        node_id = node_id or self.node_id
        port = self.conn.baremetal.create_port(node_uuid=node_id, **kwargs)
        self.addCleanup(
            lambda: self.conn.baremetal.delete_port(port.id,
                                                    ignore_missing=True))
        return port

    def create_port_group(self, node_id=None, **kwargs):
        node_id = node_id or self.node_id
        port_group = self.conn.baremetal.create_port_group(node_uuid=node_id,
                                                           **kwargs)
        self.addCleanup(
            lambda: self.conn.baremetal.delete_port_group(port_group.id,
                                                          ignore_missing=True))
        return port_group

    def create_volume_connector(self, node_id=None, **kwargs):
        node_id = node_id or self.node_id
        volume_connector = self.conn.baremetal.create_volume_connector(
            node_uuid=node_id, **kwargs)

        self.addCleanup(
            lambda:
            self.conn.baremetal.delete_volume_connector(volume_connector.id,
                                                        ignore_missing=True))
        return volume_connector

    def create_volume_target(self, node_id=None, **kwargs):
        node_id = node_id or self.node_id
        volume_target = self.conn.baremetal.create_volume_target(
            node_uuid=node_id, **kwargs)

        self.addCleanup(
            lambda:
            self.conn.baremetal.delete_volume_target(volume_target.id,
                                                     ignore_missing=True))
        return volume_target

    def create_deploy_template(self, **kwargs):
        """Create a new deploy_template from attributes.
        """

        deploy_template = self.conn.baremetal.create_deploy_template(
            **kwargs)

        self.addCleanup(
            lambda: self.conn.baremetal.delete_deploy_template(
                deploy_template.id,
                ignore_missing=True))
        return deploy_template



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.tests.functional.shared_file_system import base


class LimitTest(base.BaseSharedFileSystemTest):

    def test_limits(self):
        limits = self.user_cloud.shared_file_system.limits()
        self.assertGreater(len(list(limits)), 0)
        for limit in limits:
            for attribute in ("maxTotalReplicaGigabytes",
                              "maxTotalShares",
                              "maxTotalShareGigabytes",
                              "maxTotalShareNetworks",
                              "maxTotalShareSnapshots",
                              "maxTotalShareReplicas",
                              "maxTotalSnapshotGigabytes",
                              "totalReplicaGigabytesUsed",
                              "totalShareGigabytesUsed",
                              "totalSharesUsed",
                              "totalShareNetworksUsed",
                              "totalShareSnapshotsUsed",
                              "totalSnapshotGigabytesUsed",
                              "totalShareReplicasUsed"):
                self.assertTrue(hasattr(limit, attribute))



# Copyright (c) 2018 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import sys

import pbr.version


def show_version(args):
    print("OpenstackSDK Version %s" %
          pbr.version.VersionInfo('openstacksdk').version_string_with_vcs())


parser = argparse.ArgumentParser(description="Openstack SDK")
subparsers = parser.add_subparsers(title='commands',
                                   dest='command')

cmd_version = subparsers.add_parser('version',
                                    help='show Openstack SDK version')
cmd_version.set_defaults(func=show_version)

args = parser.parse_args()

if not args.command:
    parser.print_help()
    sys.exit(1)

args.func(args)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# import types so that we can reference ListType in sphinx param declarations.
# We can't just use list, because sphinx gets confused by
# openstack.resource.Resource.list and openstack.resource2.Resource.list
import ipaddress
# import jsonpatch
import threading
import time
import types  # noqa

from openstack.cloud import _normalize
from openstack.cloud import _utils
from openstack.cloud import exc
from openstack.cloud import meta
from openstack import exceptions
from openstack import proxy
from openstack import utils

_CONFIG_DOC_URL = (
    "https://docs.openstack.org/openstacksdk/latest/"
    "user/config/configuration.html")


class FloatingIPCloudMixin(_normalize.Normalizer):

    def __init__(self):
        self.private = self.config.config.get('private', False)

        self._floating_ip_source = self.config.config.get(
            'floating_ip_source')
        if self._floating_ip_source:
            if self._floating_ip_source.lower() == 'none':
                self._floating_ip_source = None
            else:
                self._floating_ip_source = self._floating_ip_source.lower()

        self._floating_ips = None
        self._floating_ips_time = 0
        self._floating_ips_lock = threading.Lock()

        self._floating_network_by_router = None
        self._floating_network_by_router_run = False
        self._floating_network_by_router_lock = threading.Lock()

    def search_floating_ip_pools(self, name=None, filters=None):
        pools = self.list_floating_ip_pools()
        return _utils._filter_list(pools, name, filters)

    # With Neutron, there are some cases in which full server side filtering is
    # not possible (e.g. nested attributes or list of objects) so we also need
    # to use the client-side filtering
    # The same goes for all neutron-related search/get methods!
    def search_floating_ips(self, id=None, filters=None):
        # `filters` could be a jmespath expression which Neutron server doesn't
        # understand, obviously.
        if self._use_neutron_floating() and isinstance(filters, dict):
            filter_keys = ['router_id', 'status', 'tenant_id', 'project_id',
                           'revision_number', 'description',
                           'floating_network_id', 'fixed_ip_address',
                           'floating_ip_address', 'port_id', 'sort_dir',
                           'sort_key', 'tags', 'tags-any', 'not-tags',
                           'not-tags-any', 'fields']
            neutron_filters = {k: v for k, v in filters.items()
                               if k in filter_keys}
            kwargs = {'filters': neutron_filters}
        else:
            kwargs = {}
        floating_ips = self.list_floating_ips(**kwargs)
        return _utils._filter_list(floating_ips, id, filters)

    def _neutron_list_floating_ips(self, filters=None):
        if not filters:
            filters = {}
        data = list(self.network.ips(**filters))
        return data

    def _nova_list_floating_ips(self):
        try:
            data = proxy._json_response(
                self.compute.get('/os-floating-ips'))
        except exc.OpenStackCloudURINotFound:
            return []
        return self._get_and_munchify('floating_ips', data)

    def get_floating_ip(self, id, filters=None):
        """Get a floating IP by ID

        :param id: ID of the floating IP.
        :param filters:
            A dictionary of meta data to use for further filtering. Elements
            of this dictionary may, themselves, be dictionaries. Example::

                {
                    'last_name': 'Smith',
                    'other': {
                        'gender': 'Female'
                    }
                }

            OR
            A string containing a jmespath expression for further filtering.
            Example:: "[?last_name==`Smith`] | [?other.gender]==`Female`]"

        :returns: A floating IP ``munch.Munch`` or None if no matching floating
            IP is found.

        """
        return _utils._get_entity(self, 'floating_ip', id, filters)

    def _list_floating_ips(self, filters=None):
        if self._use_neutron_floating():
            try:
                return self._normalize_floating_ips(
                    self._neutron_list_floating_ips(filters))
            except exc.OpenStackCloudURINotFound as e:
                # Nova-network don't support server-side floating ips
                # filtering, so it's safer to return and empty list than
                # to fallback to Nova which may return more results that
                # expected.
                if filters:
                    self.log.error(
                        "Neutron returned NotFound for floating IPs, which"
                        " means this cloud doesn't have neutron floating ips."
                        " shade can't fallback to trying Nova since nova"
                        " doesn't support server-side filtering when listing"
                        " floating ips and filters were given. If you do not"
                        " think shade should be attempting to list floating"
                        " ips on neutron, it is possible to control the"
                        " behavior by setting floating_ip_source to 'nova' or"
                        " None for cloud: %(cloud)s. If you are not already"
                        " using clouds.yaml to configure settings for your"
                        " cloud(s), and you want to configure this setting,"
                        " you will need a clouds.yaml file. For more"
                        " information, please see %(doc_url)s", {
                            'cloud': self.name,
                            'doc_url': _CONFIG_DOC_URL,
                        }
                    )
                    # We can't fallback to nova because we push-down filters.
                    # We got a 404 which means neutron doesn't exist. If the
                    # user
                    return []
                self.log.debug(
                    "Something went wrong talking to neutron API: "
                    "'%(msg)s'. Trying with Nova.", {'msg': str(e)})
                # Fall-through, trying with Nova
        else:
            if filters:
                raise ValueError(
                    "Nova-network don't support server-side floating ips "
                    "filtering. Use the search_floatting_ips method instead"
                )

        floating_ips = self._nova_list_floating_ips()
        return self._normalize_floating_ips(floating_ips)

    def list_floating_ip_pools(self):
        """List all available floating IP pools.

        NOTE: This function supports the nova-net view of the world. nova-net
        has been deprecated, so it's highly recommended to switch to using
        neutron. `get_external_ipv4_floating_networks` is what you should
        almost certainly be using.

        :returns: A list of floating IP pool ``munch.Munch``.

        """
        if not self._has_nova_extension('os-floating-ip-pools'):
            raise exc.OpenStackCloudUnavailableExtension(
                'Floating IP pools extension is not available on target cloud')

        data = proxy._json_response(
            self.compute.get('os-floating-ip-pools'),
            error_message="Error fetching floating IP pool list")
        pools = self._get_and_munchify('floating_ip_pools', data)
        return [{'name': p['name']} for p in pools]

    def list_floating_ips(self, filters=None):
        """List all available floating IPs.

        :param filters: (optional) dict of filter conditions to push down
        :returns: A list of floating IP ``munch.Munch``.

        """
        # If pushdown filters are specified and we do not have batched caching
        # enabled, bypass local caching and push down the filters.
        if filters and self._FLOAT_AGE == 0:
            return self._list_floating_ips(filters)

        if (time.time() - self._floating_ips_time) >= self._FLOAT_AGE:
            # Since we're using cached data anyway, we don't need to
            # have more than one thread actually submit the list
            # floating ips task.  Let the first one submit it while holding
            # a lock, and the non-blocking acquire method will cause
            # subsequent threads to just skip this and use the old
            # data until it succeeds.
            # Initially when we never got data, block to retrieve some data.
            first_run = self._floating_ips is None
            if self._floating_ips_lock.acquire(first_run):
                try:
                    if not (first_run and self._floating_ips is not None):
                        self._floating_ips = self._list_floating_ips()
                        self._floating_ips_time = time.time()
                finally:
                    self._floating_ips_lock.release()
        # Wrap the return with filter_list so that if filters were passed
        # but we were batching/caching and thus always fetching the whole
        # list from the cloud, we still return a filtered list.
        return _utils._filter_list(self._floating_ips, None, filters)

    def get_floating_ip_by_id(self, id):
        """ Get a floating ip by ID

        :param id: ID of the floating ip.
        :returns: A floating ip ``munch.Munch``.
        """
        error_message = "Error getting floating ip with ID {id}".format(id=id)

        if self._use_neutron_floating():
            fip = self.network.get_ip(id)
            return self._normalize_floating_ip(fip)
        else:
            data = proxy._json_response(
                self.compute.get('/os-floating-ips/{id}'.format(id=id)),
                error_message=error_message)
            return self._normalize_floating_ip(
                self._get_and_munchify('floating_ip', data))

    def _neutron_available_floating_ips(
            self, network=None, project_id=None, server=None):
        """Get a floating IP from a network.

        Return a list of available floating IPs or allocate a new one and
        return it in a list of 1 element.

        :param network: A single network name or ID, or a list of them.
        :param server: (server) Server the Floating IP is for

        :returns: a list of floating IP addresses.

        :raises: ``OpenStackCloudResourceNotFound``, if an external network
            that meets the specified criteria cannot be found.
        """
        if project_id is None:
            # Make sure we are only listing floatingIPs allocated the current
            # tenant. This is the default behaviour of Nova
            project_id = self.current_project_id

        if network:
            if isinstance(network, str):
                network = [network]

            # Use given list to get first matching external network
            floating_network_id = None
            for net in network:
                for ext_net in self.get_external_ipv4_floating_networks():
                    if net in (ext_net['name'], ext_net['id']):
                        floating_network_id = ext_net['id']
                        break
                if floating_network_id:
                    break

            if floating_network_id is None:
                raise exc.OpenStackCloudResourceNotFound(
                    "unable to find external network {net}".format(
                        net=network)
                )
        else:
            floating_network_id = self._get_floating_network_id()

        filters = {
            'port': None,
            'network': floating_network_id,
            'location': {'project': {'id': project_id}},
        }

        floating_ips = self._list_floating_ips()
        available_ips = _utils._filter_list(
            floating_ips, name_or_id=None, filters=filters)
        if available_ips:
            return available_ips

        # No available IP found or we didn't try
        # allocate a new Floating IP
        f_ip = self._neutron_create_floating_ip(
            network_id=floating_network_id, server=server)

        return [f_ip]

    def _nova_available_floating_ips(self, pool=None):
        """Get available floating IPs from a floating IP pool.

        Return a list of available floating IPs or allocate a new one and
        return it in a list of 1 element.

        :param pool: Nova floating IP pool name.

        :returns: a list of floating IP addresses.

        :raises: ``OpenStackCloudResourceNotFound``, if a floating IP pool
            is not specified and cannot be found.
        """

        with _utils.shade_exceptions(
                "Unable to create floating IP in pool {pool}".format(
                    pool=pool)):
            if pool is None:
                pools = self.list_floating_ip_pools()
                if not pools:
                    raise exc.OpenStackCloudResourceNotFound(
                        "unable to find a floating ip pool")
                pool = pools[0]['name']

            filters = {
                'instance_id': None,
                'pool': pool
            }

            floating_ips = self._nova_list_floating_ips()
            available_ips = _utils._filter_list(
                floating_ips, name_or_id=None, filters=filters)
            if available_ips:
                return available_ips

            # No available IP found or we did not try.
            # Allocate a new Floating IP
            f_ip = self._nova_create_floating_ip(pool=pool)

            return [f_ip]

    def _find_floating_network_by_router(self):
        """Find the network providing floating ips by looking at routers."""

        if self._floating_network_by_router_lock.acquire(
                not self._floating_network_by_router_run):
            if self._floating_network_by_router_run:
                self._floating_network_by_router_lock.release()
                return self._floating_network_by_router
            try:
                for router in self.list_routers():
                    if router['admin_state_up']:
                        network_id = router.get(
                            'external_gateway_info', {}).get('network_id')
                        if network_id:
                            self._floating_network_by_router = network_id
            finally:
                self._floating_network_by_router_run = True
                self._floating_network_by_router_lock.release()
        return self._floating_network_by_router

    def available_floating_ip(self, network=None, server=None):
        """Get a floating IP from a network or a pool.

        Return the first available floating IP or allocate a new one.

        :param network: Name or ID of the network.
        :param server: Server the IP is for if known

        :returns: a (normalized) structure with a floating IP address
                  description.
        """
        if self._use_neutron_floating():
            try:
                f_ips = self._normalize_floating_ips(
                    self._neutron_available_floating_ips(
                        network=network, server=server))
                return f_ips[0]
            except exc.OpenStackCloudURINotFound as e:
                self.log.debug(
                    "Something went wrong talking to neutron API: "
                    "'%(msg)s'. Trying with Nova.", {'msg': str(e)})
                # Fall-through, trying with Nova

        f_ips = self._normalize_floating_ips(
            self._nova_available_floating_ips(pool=network)
        )
        return f_ips[0]

    def _get_floating_network_id(self):
        # Get first existing external IPv4 network
        networks = self.get_external_ipv4_floating_networks()
        if networks:
            floating_network_id = networks[0]['id']
        else:
            floating_network = self._find_floating_network_by_router()
            if floating_network:
                floating_network_id = floating_network
            else:
                raise exc.OpenStackCloudResourceNotFound(
                    "unable to find an external network")
        return floating_network_id

    def create_floating_ip(self, network=None, server=None,
                           fixed_address=None, nat_destination=None,
                           port=None, wait=False, timeout=60):
        """Allocate a new floating IP from a network or a pool.

        :param network: Name or ID of the network
                        that the floating IP should come from.
        :param server: (optional) Server dict for the server to create
                       the IP for and to which it should be attached.
        :param fixed_address: (optional) Fixed IP to attach the floating
                              ip to.
        :param nat_destination: (optional) Name or ID of the network
                                that the fixed IP to attach the floating
                                IP to should be on.
        :param port: (optional) The port ID that the floating IP should be
                                attached to. Specifying a port conflicts
                                with specifying a server, fixed_address or
                                nat_destination.
        :param wait: (optional) Whether to wait for the IP to be active.
                     Defaults to False. Only applies if a server is
                     provided.
        :param timeout: (optional) How long to wait for the IP to be active.
                        Defaults to 60. Only applies if a server is
                        provided.

        :returns: a floating IP address

        :raises: ``OpenStackCloudException``, on operation error.
        """
        if self._use_neutron_floating():
            try:
                return self._neutron_create_floating_ip(
                    network_name_or_id=network, server=server,
                    fixed_address=fixed_address,
                    nat_destination=nat_destination,
                    port=port,
                    wait=wait, timeout=timeout)
            except exc.OpenStackCloudURINotFound as e:
                self.log.debug(
                    "Something went wrong talking to neutron API: "
                    "'%(msg)s'. Trying with Nova.", {'msg': str(e)})
                # Fall-through, trying with Nova

        if port:
            raise exc.OpenStackCloudException(
                "This cloud uses nova-network which does not support"
                " arbitrary floating-ip/port mappings. Please nudge"
                " your cloud provider to upgrade the networking stack"
                " to neutron, or alternately provide the server,"
                " fixed_address and nat_destination arguments as appropriate")
        # Else, we are using Nova network
        f_ips = self._normalize_floating_ips(
            [self._nova_create_floating_ip(pool=network)])
        return f_ips[0]

    def _submit_create_fip(self, kwargs):
        # Split into a method to aid in test mocking
        data = self.network.create_ip(**kwargs)
        return self._normalize_floating_ip(data)

    def _neutron_create_floating_ip(
            self, network_name_or_id=None, server=None,
            fixed_address=None, nat_destination=None,
            port=None,
            wait=False, timeout=60, network_id=None):

        if not network_id:
            if network_name_or_id:
                try:
                    network = self.network.find_network(network_name_or_id)
                except exceptions.ResourceNotFound:
                    raise exc.OpenStackCloudResourceNotFound(
                        "unable to find network for floating ips with ID "
                        "{0}".format(network_name_or_id))
                network_id = network['id']
            else:
                network_id = self._get_floating_network_id()
        kwargs = {
            'floating_network_id': network_id,
        }
        if not port:
            if server:
                (port_obj, fixed_ip_address) = self._nat_destination_port(
                    server, fixed_address=fixed_address,
                    nat_destination=nat_destination)
                if port_obj:
                    port = port_obj['id']
                if fixed_ip_address:
                    kwargs['fixed_ip_address'] = fixed_ip_address
        if port:
            kwargs['port_id'] = port

        fip = self._submit_create_fip(kwargs)
        fip_id = fip['id']

        if port:
            # The FIP is only going to become active in this context
            # when we've attached it to something, which only occurs
            # if we've provided a port as a parameter
            if wait:
                try:
                    for count in utils.iterate_timeout(
                            timeout,
                            "Timeout waiting for the floating IP"
                            " to be ACTIVE",
                            wait=self._FLOAT_AGE):
                        fip = self.get_floating_ip(fip_id)
                        if fip and fip['status'] == 'ACTIVE':
                            break
                except exc.OpenStackCloudTimeout:
                    self.log.error(
                        "Timed out on floating ip %(fip)s becoming active."
                        " Deleting", {'fip': fip_id})
                    try:
                        self.delete_floating_ip(fip_id)
                    except Exception as e:
                        self.log.error(
                            "FIP LEAK: Attempted to delete floating ip "
                            "%(fip)s but received %(exc)s exception: "
                            "%(err)s", {'fip': fip_id, 'exc': e.__class__,
                                        'err': str(e)})
                    raise
            if fip['port_id'] != port:
                if server:
                    raise exc.OpenStackCloudException(
                        "Attempted to create FIP on port {port} for server"
                        " {server} but FIP has port {port_id}".format(
                            port=port, port_id=fip['port_id'],
                            server=server['id']))
                else:
                    raise exc.OpenStackCloudException(
                        "Attempted to create FIP on port {port}"
                        " but something went wrong".format(port=port))
        return fip

    def _nova_create_floating_ip(self, pool=None):
        with _utils.shade_exceptions(
                "Unable to create floating IP in pool {pool}".format(
                    pool=pool)):
            if pool is None:
                pools = self.list_floating_ip_pools()
                if not pools:
                    raise exc.OpenStackCloudResourceNotFound(
                        "unable to find a floating ip pool")
                pool = pools[0]['name']

            data = proxy._json_response(self.compute.post(
                '/os-floating-ips', json=dict(pool=pool)))
            pool_ip = self._get_and_munchify('floating_ip', data)
            # TODO(mordred) Remove this - it's just for compat
            data = proxy._json_response(
                self.compute.get('/os-floating-ips/{id}'.format(
                    id=pool_ip['id'])))
            return self._get_and_munchify('floating_ip', data)

    def delete_floating_ip(self, floating_ip_id, retry=1):
        """Deallocate a floating IP from a project.

        :param floating_ip_id: a floating IP address ID.
        :param retry: number of times to retry. Optional, defaults to 1,
                      which is in addition to the initial delete call.
                      A value of 0 will also cause no checking of results to
                      occur.

        :returns: True if the IP address has been deleted, False if the IP
                  address was not found.

        :raises: ``OpenStackCloudException``, on operation error.
        """
        for count in range(0, max(0, retry) + 1):
            result = self._delete_floating_ip(floating_ip_id)

            if (retry == 0) or not result:
                return result

            # Wait for the cached floating ip list to be regenerated
            if self._FLOAT_AGE:
                time.sleep(self._FLOAT_AGE)

            # neutron sometimes returns success when deleting a floating
            # ip. That's awesome. SO - verify that the delete actually
            # worked. Some clouds will set the status to DOWN rather than
            # deleting the IP immediately. This is, of course, a bit absurd.
            f_ip = self.get_floating_ip(id=floating_ip_id)
            if not f_ip or f_ip['status'] == 'DOWN':
                return True

        raise exc.OpenStackCloudException(
            "Attempted to delete Floating IP {ip} with ID {id} a total of"
            " {retry} times. Although the cloud did not indicate any errors"
            " the floating ip is still in existence. Aborting further"
            " operations.".format(
                id=floating_ip_id, ip=f_ip['floating_ip_address'],
                retry=retry + 1))

    def _delete_floating_ip(self, floating_ip_id):
        if self._use_neutron_floating():
            try:
                return self._neutron_delete_floating_ip(floating_ip_id)
            except exc.OpenStackCloudURINotFound as e:
                self.log.debug(
                    "Something went wrong talking to neutron API: "
                    "'%(msg)s'. Trying with Nova.", {'msg': str(e)})
        return self._nova_delete_floating_ip(floating_ip_id)

    def _neutron_delete_floating_ip(self, floating_ip_id):
        try:
            self.network.delete_ip(
                floating_ip_id, ignore_missing=False
            )
        except exceptions.ResourceNotFound:
            return False
        return True

    def _nova_delete_floating_ip(self, floating_ip_id):
        try:
            proxy._json_response(
                self.compute.delete(
                    '/os-floating-ips/{id}'.format(id=floating_ip_id)),
                error_message='Unable to delete floating IP {fip_id}'.format(
                    fip_id=floating_ip_id))
        except exc.OpenStackCloudURINotFound:
            return False
        return True

    def delete_unattached_floating_ips(self, retry=1):
        """Safely delete unattached floating ips.

        If the cloud can safely purge any unattached floating ips without
        race conditions, do so.

        Safely here means a specific thing. It means that you are not running
        this while another process that might do a two step create/attach
        is running. You can safely run this  method while another process
        is creating servers and attaching floating IPs to them if either that
        process is using add_auto_ip from shade, or is creating the floating
        IPs by passing in a server to the create_floating_ip call.

        :param retry: number of times to retry. Optional, defaults to 1,
                      which is in addition to the initial delete call.
                      A value of 0 will also cause no checking of results to
                      occur.

        :returns: Number of Floating IPs deleted, False if none

        :raises: ``OpenStackCloudException``, on operation error.
        """
        processed = []
        if self._use_neutron_floating():
            for ip in self.list_floating_ips():
                if not ip['attached']:
                    processed.append(self.delete_floating_ip(
                        floating_ip_id=ip['id'], retry=retry))
        return len(processed) if all(processed) else False

    def _attach_ip_to_server(
            self, server, floating_ip,
            fixed_address=None, wait=False,
            timeout=60, skip_attach=False, nat_destination=None):
        """Attach a floating IP to a server.

        :param server: Server dict
        :param floating_ip: Floating IP dict to attach
        :param fixed_address: (optional) fixed address to which attach the
                              floating IP to.
        :param wait: (optional) Wait for the address to appear as assigned
                     to the server. Defaults to False.
        :param timeout: (optional) Seconds to wait, defaults to 60.
                        See the ``wait`` parameter.
        :param skip_attach: (optional) Skip the actual attach and just do
                            the wait. Defaults to False.
        :param nat_destination: The fixed network the server's port for the
                                FIP to attach to will come from.

        :returns: The server ``munch.Munch``

        :raises: OpenStackCloudException, on operation error.
        """
        # Short circuit if we're asking to attach an IP that's already
        # attached
        ext_ip = meta.get_server_ip(server, ext_tag='floating', public=True)
        if not ext_ip and floating_ip['port_id']:
            # When we came here from reuse_fip and created FIP it might be
            # already attached, but the server info might be also
            # old to check whether it belongs to us now, thus refresh
            # the server data and try again. There are some clouds, which
            # explicitely forbids FIP assign call if it is already assigned.
            server = self.get_server_by_id(server['id'])
            ext_ip = meta.get_server_ip(server, ext_tag='floating',
                                        public=True)
        if ext_ip == floating_ip['floating_ip_address']:
            return server

        if self._use_neutron_floating():
            if not skip_attach:
                try:
                    self._neutron_attach_ip_to_server(
                        server=server, floating_ip=floating_ip,
                        fixed_address=fixed_address,
                        nat_destination=nat_destination)
                except exc.OpenStackCloudURINotFound as e:
                    self.log.debug(
                        "Something went wrong talking to neutron API: "
                        "'%(msg)s'. Trying with Nova.", {'msg': str(e)})
                    # Fall-through, trying with Nova
        else:
            # Nova network
            self._nova_attach_ip_to_server(
                server_id=server['id'], floating_ip_id=floating_ip['id'],
                fixed_address=fixed_address)

        if wait:
            # Wait for the address to be assigned to the server
            server_id = server['id']
            for _ in utils.iterate_timeout(
                    timeout,
                    "Timeout waiting for the floating IP to be attached.",
                    wait=self._SERVER_AGE):
                server = self.get_server_by_id(server_id)
                ext_ip = meta.get_server_ip(
                    server, ext_tag='floating', public=True)
                if ext_ip == floating_ip['floating_ip_address']:
                    return server
        return server

    def _neutron_attach_ip_to_server(
            self, server, floating_ip, fixed_address=None,
            nat_destination=None):

        # Find an available port
        (port, fixed_address) = self._nat_destination_port(
            server, fixed_address=fixed_address,
            nat_destination=nat_destination)
        if not port:
            raise exc.OpenStackCloudException(
                "unable to find a port for server {0}".format(
                    server['id']))

        floating_ip_args = {'port_id': port['id']}
        if fixed_address is not None:
            floating_ip_args['fixed_ip_address'] = fixed_address

        return self.network.update_ip(
            floating_ip,
            **floating_ip_args)

    def _nova_attach_ip_to_server(self, server_id, floating_ip_id,
                                  fixed_address=None):
        f_ip = self.get_floating_ip(
            id=floating_ip_id)
        if f_ip is None:
            raise exc.OpenStackCloudException(
                "unable to find floating IP {0}".format(floating_ip_id))
        error_message = "Error attaching IP {ip} to instance {id}".format(
            ip=floating_ip_id, id=server_id)
        body = {
            'address': f_ip['floating_ip_address']
        }
        if fixed_address:
            body['fixed_address'] = fixed_address
        return proxy._json_response(
            self.compute.post(
                '/servers/{server_id}/action'.format(server_id=server_id),
                json=dict(addFloatingIp=body)),
            error_message=error_message)

    def detach_ip_from_server(self, server_id, floating_ip_id):
        """Detach a floating IP from a server.

        :param server_id: ID of a server.
        :param floating_ip_id: Id of the floating IP to detach.

        :returns: True if the IP has been detached, or False if the IP wasn't
                  attached to any server.

        :raises: ``OpenStackCloudException``, on operation error.
        """
        if self._use_neutron_floating():
            try:
                return self._neutron_detach_ip_from_server(
                    server_id=server_id, floating_ip_id=floating_ip_id)
            except exc.OpenStackCloudURINotFound as e:
                self.log.debug(
                    "Something went wrong talking to neutron API: "
                    "'%(msg)s'. Trying with Nova.", {'msg': str(e)})
                # Fall-through, trying with Nova

        # Nova network
        self._nova_detach_ip_from_server(
            server_id=server_id, floating_ip_id=floating_ip_id)

    def _neutron_detach_ip_from_server(self, server_id, floating_ip_id):
        f_ip = self.get_floating_ip(id=floating_ip_id)
        if f_ip is None or not f_ip['attached']:
            return False
        try:
            self.network.update_ip(
                floating_ip_id,
                port_id=None
            )
        except exceptions.SDKException:
            raise exceptions.SDKException(
                ("Error detaching IP {ip} from "
                 "server {server_id}".format(
                     ip=floating_ip_id, server_id=server_id)))

        return True

    def _nova_detach_ip_from_server(self, server_id, floating_ip_id):

        f_ip = self.get_floating_ip(id=floating_ip_id)
        if f_ip is None:
            raise exc.OpenStackCloudException(
                "unable to find floating IP {0}".format(floating_ip_id))
        error_message = "Error detaching IP {ip} from instance {id}".format(
            ip=floating_ip_id, id=server_id)
        return proxy._json_response(
            self.compute.post(
                '/servers/{server_id}/action'.format(server_id=server_id),
                json=dict(removeFloatingIp=dict(
                    address=f_ip['floating_ip_address']))),
            error_message=error_message)

        return True

    def _add_ip_from_pool(
            self, server, network, fixed_address=None, reuse=True,
            wait=False, timeout=60, nat_destination=None):
        """Add a floating IP to a server from a given pool

        This method reuses available IPs, when possible, or allocate new IPs
        to the current tenant.
        The floating IP is attached to the given fixed address or to the
        first server port/fixed address

        :param server: Server dict
        :param network: Name or ID of the network.
        :param fixed_address: a fixed address
        :param reuse: Try to reuse existing ips. Defaults to True.
        :param wait: (optional) Wait for the address to appear as assigned
                     to the server. Defaults to False.
        :param timeout: (optional) Seconds to wait, defaults to 60.
                        See the ``wait`` parameter.
        :param nat_destination: (optional) the name of the network of the
                                port to associate with the floating ip.

        :returns: the updated server ``munch.Munch``
        """
        if reuse:
            f_ip = self.available_floating_ip(network=network)
        else:
            start_time = time.time()
            f_ip = self.create_floating_ip(
                server=server,
                network=network, nat_destination=nat_destination,
                wait=wait, timeout=timeout)
            timeout = timeout - (time.time() - start_time)
            # Wait for cache invalidation time so that we don't try
            # to attach the FIP a second time below
            time.sleep(self._SERVER_AGE)
            server = self.get_server(server.id)

        # We run attach as a second call rather than in the create call
        # because there are code flows where we will not have an attached
        # FIP yet. However, even if it was attached in the create, we run
        # the attach function below to get back the server dict refreshed
        # with the FIP information.
        return self._attach_ip_to_server(
            server=server, floating_ip=f_ip, fixed_address=fixed_address,
            wait=wait, timeout=timeout, nat_destination=nat_destination)

    def add_ip_list(
            self, server, ips, wait=False, timeout=60,
            fixed_address=None):
        """Attach a list of IPs to a server.

        :param server: a server object
        :param ips: list of floating IP addresses or a single address
        :param wait: (optional) Wait for the address to appear as assigned
                     to the server. Defaults to False.
        :param timeout: (optional) Seconds to wait, defaults to 60.
                        See the ``wait`` parameter.
        :param fixed_address: (optional) Fixed address of the server to
                                         attach the IP to

        :returns: The updated server ``munch.Munch``

        :raises: ``OpenStackCloudException``, on operation error.
        """

        if type(ips) != list:
            ips = [ips]

        for ip in ips:
            f_ip = self.get_floating_ip(
                id=None, filters={'floating_ip_address': ip})
            server = self._attach_ip_to_server(
                server=server, floating_ip=f_ip, wait=wait, timeout=timeout,
                fixed_address=fixed_address)
        return server

    def add_auto_ip(self, server, wait=False, timeout=60, reuse=True):
        """Add a floating IP to a server.

        This method is intended for basic usage. For advanced network
        architecture (e.g. multiple external networks or servers with multiple
        interfaces), use other floating IP methods.

        This method can reuse available IPs, or allocate new IPs to the current
        project.

        :param server: a server dictionary.
        :param reuse: Whether or not to attempt to reuse IPs, defaults
                      to True.
        :param wait: (optional) Wait for the address to appear as assigned
                     to the server. Defaults to False.
        :param timeout: (optional) Seconds to wait, defaults to 60.
                        See the ``wait`` parameter.
        :param reuse: Try to reuse existing ips. Defaults to True.

        :returns: Floating IP address attached to server.

        """
        server = self._add_auto_ip(
            server, wait=wait, timeout=timeout, reuse=reuse)
        return server['interface_ip'] or None

    def _add_auto_ip(self, server, wait=False, timeout=60, reuse=True):
        skip_attach = False
        created = False
        if reuse:
            f_ip = self.available_floating_ip(server=server)
        else:
            start_time = time.time()
            f_ip = self.create_floating_ip(
                server=server, wait=wait, timeout=timeout)
            timeout = timeout - (time.time() - start_time)
            if server:
                # This gets passed in for both nova and neutron
                # but is only meaningful for the neutron logic branch
                skip_attach = True
            created = True

        try:
            # We run attach as a second call rather than in the create call
            # because there are code flows where we will not have an attached
            # FIP yet. However, even if it was attached in the create, we run
            # the attach function below to get back the server dict refreshed
            # with the FIP information.
            return self._attach_ip_to_server(
                server=server, floating_ip=f_ip, wait=wait, timeout=timeout,
                skip_attach=skip_attach)
        except exc.OpenStackCloudTimeout:
            if self._use_neutron_floating() and created:
                # We are here because we created an IP on the port
                # It failed. Delete so as not to leak an unmanaged
                # resource
                self.log.error(
                    "Timeout waiting for floating IP to become"
                    " active. Floating IP %(ip)s:%(id)s was created for"
                    " server %(server)s but is being deleted due to"
                    " activation failure.", {
                        'ip': f_ip['floating_ip_address'],
                        'id': f_ip['id'],
                        'server': server['id']})
                try:
                    self.delete_floating_ip(f_ip['id'])
                except Exception as e:
                    self.log.error(
                        "FIP LEAK: Attempted to delete floating ip "
                        "%(fip)s but received %(exc)s exception: %(err)s",
                        {'fip': f_ip['id'], 'exc': e.__class__, 'err': str(e)})
                    raise e
            raise

    def add_ips_to_server(
            self, server, auto_ip=True, ips=None, ip_pool=None,
            wait=False, timeout=60, reuse=True, fixed_address=None,
            nat_destination=None):
        if ip_pool:
            server = self._add_ip_from_pool(
                server, ip_pool, reuse=reuse, wait=wait, timeout=timeout,
                fixed_address=fixed_address, nat_destination=nat_destination)
        elif ips:
            server = self.add_ip_list(
                server, ips, wait=wait, timeout=timeout,
                fixed_address=fixed_address)
        elif auto_ip:
            if self._needs_floating_ip(server, nat_destination):
                server = self._add_auto_ip(
                    server, wait=wait, timeout=timeout, reuse=reuse)
        return server

    def _needs_floating_ip(self, server, nat_destination):
        """Figure out if auto_ip should add a floating ip to this server.

        If the server has a floating ip it does not need another one.

        If the server does not have a fixed ip address it does not need a
        floating ip.

        If self.private then the server does not need a floating ip.

        If the cloud runs nova, and the server has a private address and not a
        public address, then the server needs a floating ip.

        If the server has a fixed ip address and no floating ip address and the
        cloud has a network from which floating IPs come that is connected via
        a router to the network from which the fixed ip address came,
        then the server needs a floating ip.

        If the server has a fixed ip address and no floating ip address and the
        cloud does not have a network from which floating ips come, or it has
        one but that network is not connected to the network from which
        the server's fixed ip address came via a router, then the
        server does not need a floating ip.
        """
        if not self._has_floating_ips():
            return False

        if server['addresses'] is None:
            # fetch missing server details, e.g. because
            # meta.add_server_interfaces() was not called
            server = self.compute.get_server(server)

        if server['public_v4'] \
           or any([any([address['OS-EXT-IPS:type'] == 'floating'
                        for address in addresses])
                   for addresses
                   in (server['addresses'] or {}).values()]):
            return False

        if not server['private_v4'] \
           and not any([any([address['OS-EXT-IPS:type'] == 'fixed'
                             for address in addresses])
                        for addresses
                        in (server['addresses'] or {}).values()]):
            return False

        if self.private:
            return False

        if not self.has_service('network'):
            return True

        # No floating ip network - no FIPs
        try:
            self._get_floating_network_id()
        except exc.OpenStackCloudException:
            return False

        (port_obj, fixed_ip_address) = self._nat_destination_port(
            server, nat_destination=nat_destination)

        if not port_obj or not fixed_ip_address:
            return False

        return True

    def _nat_destination_port(
            self, server, fixed_address=None, nat_destination=None):
        """Returns server port that is on a nat_destination network

        Find a port attached to the server which is on a network which
        has a subnet which can be the destination of NAT. Such a network
        is referred to in shade as a "nat_destination" network. So this
        then is a function which returns a port on such a network that is
        associated with the given server.

        :param server: Server dict.
        :param fixed_address: Fixed ip address of the port
        :param nat_destination: Name or ID of the network of the port.
        """
        # If we are caching port lists, we may not find the port for
        # our server if the list is old.  Try for at least 2 cache
        # periods if that is the case.
        if self._PORT_AGE:
            timeout = self._PORT_AGE * 2
        else:
            timeout = None
        for count in utils.iterate_timeout(
                timeout,
                "Timeout waiting for port to show up in list",
                wait=self._PORT_AGE):
            try:
                port_filter = {'device_id': server['id']}
                ports = self.search_ports(filters=port_filter)
                break
            except exc.OpenStackCloudTimeout:
                ports = None
        if not ports:
            return (None, None)
        port = None
        if not fixed_address:
            if len(ports) > 1:
                if nat_destination:
                    nat_network = self.get_network(nat_destination)
                    if not nat_network:
                        raise exc.OpenStackCloudException(
                            'NAT Destination {nat_destination} was configured'
                            ' but not found on the cloud. Please check your'
                            ' config and your cloud and try again.'.format(
                                nat_destination=nat_destination))
                else:
                    nat_network = self.get_nat_destination()

                if not nat_network:
                    raise exc.OpenStackCloudException(
                        'Multiple ports were found for server {server}'
                        ' but none of the networks are a valid NAT'
                        ' destination, so it is impossible to add a'
                        ' floating IP. If you have a network that is a valid'
                        ' destination for NAT and we could not find it,'
                        ' please file a bug. But also configure the'
                        ' nat_destination property of the networks list in'
                        ' your clouds.yaml file. If you do not have a'
                        ' clouds.yaml file, please make one - your setup'
                        ' is complicated.'.format(server=server['id']))

                maybe_ports = []
                for maybe_port in ports:
                    if maybe_port['network_id'] == nat_network['id']:
                        maybe_ports.append(maybe_port)
                if not maybe_ports:
                    raise exc.OpenStackCloudException(
                        'No port on server {server} was found matching'
                        ' your NAT destination network {dest}. Please '
                        ' check your config'.format(
                            server=server['id'], dest=nat_network['name']))
                ports = maybe_ports

            # Select the most recent available IPv4 address
            # To do this, sort the ports in reverse order by the created_at
            # field which is a string containing an ISO DateTime (which
            # thankfully sort properly) This way the most recent port created,
            # if there are more than one, will be the arbitrary port we
            # select.
            for port in sorted(
                    ports,
                    key=lambda p: p.get('created_at', 0),
                    reverse=True):
                for address in port.get('fixed_ips', list()):
                    try:
                        ip = ipaddress.ip_address(address['ip_address'])
                    except Exception:
                        continue
                    if ip.version == 4:
                        fixed_address = address['ip_address']
                        return port, fixed_address
            raise exc.OpenStackCloudException(
                "unable to find a free fixed IPv4 address for server "
                "{0}".format(server['id']))
        # unfortunately a port can have more than one fixed IP:
        # we can't use the search_ports filtering for fixed_address as
        # they are contained in a list. e.g.
        #
        #   "fixed_ips": [
        #     {
        #       "subnet_id": "008ba151-0b8c-4a67-98b5-0d2b87666062",
        #       "ip_address": "172.24.4.2"
        #     }
        #   ]
        #
        # Search fixed_address
        for p in ports:
            for fixed_ip in p['fixed_ips']:
                if fixed_address == fixed_ip['ip_address']:
                    return (p, fixed_address)
        return (None, None)

    def _has_floating_ips(self):
        if not self._floating_ip_source:
            return False
        else:
            return self._floating_ip_source in ('nova', 'neutron')

    def _use_neutron_floating(self):
        return (self.has_service('network')
                and self._floating_ip_source == 'neutron')



# Copyright (c) 2014 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json
import os
import threading

_json_path = os.path.join(
    os.path.dirname(os.path.realpath(__file__)), 'defaults.json')
_defaults = None
_defaults_lock = threading.Lock()


# json_path argument is there for os-client-config
def get_defaults(json_path=_json_path):
    global _defaults
    if _defaults is not None:
        return _defaults.copy()
    with _defaults_lock:
        if _defaults is not None:
            # Did someone else just finish filling it?
            return _defaults.copy()
        # Python language specific defaults
        # These are defaults related to use of python libraries, they are
        # not qualities of a cloud.
        #
        # NOTE(harlowja): update a in-memory dict, before updating
        # the global one so that other callers of get_defaults do not
        # see the partially filled one.
        tmp_defaults = dict(
            api_timeout=None,
            verify=True,
            cacert=None,
            cert=None,
            key=None,
        )
        with open(json_path, 'r') as json_file:
            updates = json.load(json_file)
            if updates is not None:
                tmp_defaults.update(updates)
        _defaults = tmp_defaults
        return tmp_defaults.copy()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import uuid

from openstack.identity.v3 import _proxy
from openstack.identity.v3 import credential
from openstack.identity.v3 import domain
from openstack.identity.v3 import endpoint
from openstack.identity.v3 import group
from openstack.identity.v3 import policy
from openstack.identity.v3 import project
from openstack.identity.v3 import region
from openstack.identity.v3 import role
from openstack.identity.v3 import service
from openstack.identity.v3 import trust
from openstack.identity.v3 import user
from openstack.tests.unit import test_proxy_base

USER_ID = 'user-id-' + uuid.uuid4().hex


class TestIdentityProxyBase(test_proxy_base.TestProxyBase):
    def setUp(self):
        super(TestIdentityProxyBase, self).setUp()
        self.proxy = _proxy.Proxy(self.session)


class TestIdentityProxyCredential(TestIdentityProxyBase):

    def test_credential_create_attrs(self):
        self.verify_create(self.proxy.create_credential,
                           credential.Credential)

    def test_credential_delete(self):
        self.verify_delete(self.proxy.delete_credential,
                           credential.Credential, False)

    def test_credential_delete_ignore(self):
        self.verify_delete(self.proxy.delete_credential,
                           credential.Credential, True)

    def test_credential_find(self):
        self.verify_find(self.proxy.find_credential, credential.Credential)

    def test_credential_get(self):
        self.verify_get(self.proxy.get_credential, credential.Credential)

    def test_credentials(self):
        self.verify_list(self.proxy.credentials, credential.Credential)

    def test_credential_update(self):
        self.verify_update(self.proxy.update_credential, credential.Credential)


class TestIdentityProxyDomain(TestIdentityProxyBase):

    def test_domain_create_attrs(self):
        self.verify_create(self.proxy.create_domain, domain.Domain)

    def test_domain_delete(self):
        self.verify_delete(self.proxy.delete_domain, domain.Domain, False)

    def test_domain_delete_ignore(self):
        self.verify_delete(self.proxy.delete_domain, domain.Domain, True)

    def test_domain_find(self):
        self.verify_find(self.proxy.find_domain, domain.Domain)

    def test_domain_get(self):
        self.verify_get(self.proxy.get_domain, domain.Domain)

    def test_domains(self):
        self.verify_list(self.proxy.domains, domain.Domain)

    def test_domain_update(self):
        self.verify_update(self.proxy.update_domain, domain.Domain)


class TestIdentityProxyEndpoint(TestIdentityProxyBase):

    def test_endpoint_create_attrs(self):
        self.verify_create(self.proxy.create_endpoint, endpoint.Endpoint)

    def test_endpoint_delete(self):
        self.verify_delete(self.proxy.delete_endpoint,
                           endpoint.Endpoint, False)

    def test_endpoint_delete_ignore(self):
        self.verify_delete(self.proxy.delete_endpoint,
                           endpoint.Endpoint, True)

    def test_endpoint_find(self):
        self.verify_find(self.proxy.find_endpoint, endpoint.Endpoint)

    def test_endpoint_get(self):
        self.verify_get(self.proxy.get_endpoint, endpoint.Endpoint)

    def test_endpoints(self):
        self.verify_list(self.proxy.endpoints, endpoint.Endpoint)

    def test_endpoint_update(self):
        self.verify_update(self.proxy.update_endpoint, endpoint.Endpoint)


class TestIdentityProxyGroup(TestIdentityProxyBase):

    def test_group_create_attrs(self):
        self.verify_create(self.proxy.create_group, group.Group)

    def test_group_delete(self):
        self.verify_delete(self.proxy.delete_group, group.Group, False)

    def test_group_delete_ignore(self):
        self.verify_delete(self.proxy.delete_group, group.Group, True)

    def test_group_find(self):
        self.verify_find(self.proxy.find_group, group.Group)

    def test_group_get(self):
        self.verify_get(self.proxy.get_group, group.Group)

    def test_groups(self):
        self.verify_list(self.proxy.groups, group.Group)

    def test_group_update(self):
        self.verify_update(self.proxy.update_group, group.Group)

    def test_add_user_to_group(self):
        self._verify(
            "openstack.identity.v3.group.Group.add_user",
            self.proxy.add_user_to_group,
            method_args=['uid', 'gid'],
            expected_args=[
                self.proxy,
                self.proxy._get_resource(user.User, 'uid'),
            ]
        )

    def test_remove_user_from_group(self):
        self._verify(
            "openstack.identity.v3.group.Group.remove_user",
            self.proxy.remove_user_from_group,
            method_args=['uid', 'gid'],
            expected_args=[
                self.proxy,
                self.proxy._get_resource(user.User, 'uid'),
            ]
        )

    def test_check_user_in_group(self):
        self._verify(
            "openstack.identity.v3.group.Group.check_user",
            self.proxy.check_user_in_group,
            method_args=['uid', 'gid'],
            expected_args=[
                self.proxy,
                self.proxy._get_resource(user.User, 'uid'),
            ]
        )


class TestIdentityProxyPolicy(TestIdentityProxyBase):

    def test_policy_create_attrs(self):
        self.verify_create(self.proxy.create_policy, policy.Policy)

    def test_policy_delete(self):
        self.verify_delete(self.proxy.delete_policy, policy.Policy, False)

    def test_policy_delete_ignore(self):
        self.verify_delete(self.proxy.delete_policy, policy.Policy, True)

    def test_policy_find(self):
        self.verify_find(self.proxy.find_policy, policy.Policy)

    def test_policy_get(self):
        self.verify_get(self.proxy.get_policy, policy.Policy)

    def test_policies(self):
        self.verify_list(self.proxy.policies, policy.Policy)

    def test_policy_update(self):
        self.verify_update(self.proxy.update_policy, policy.Policy)


class TestIdentityProxyProject(TestIdentityProxyBase):

    def test_project_create_attrs(self):
        self.verify_create(self.proxy.create_project, project.Project)

    def test_project_delete(self):
        self.verify_delete(self.proxy.delete_project, project.Project, False)

    def test_project_delete_ignore(self):
        self.verify_delete(self.proxy.delete_project, project.Project, True)

    def test_project_find(self):
        self.verify_find(self.proxy.find_project, project.Project)

    def test_project_get(self):
        self.verify_get(self.proxy.get_project, project.Project)

    def test_projects(self):
        self.verify_list(self.proxy.projects, project.Project)

    def test_user_projects(self):
        self.verify_list(
            self.proxy.user_projects,
            project.UserProject,
            method_kwargs={'user': USER_ID},
            expected_kwargs={'user_id': USER_ID}
        )

    def test_project_update(self):
        self.verify_update(self.proxy.update_project, project.Project)


class TestIdentityProxyService(TestIdentityProxyBase):

    def test_service_create_attrs(self):
        self.verify_create(self.proxy.create_service, service.Service)

    def test_service_delete(self):
        self.verify_delete(self.proxy.delete_service, service.Service, False)

    def test_service_delete_ignore(self):
        self.verify_delete(self.proxy.delete_service, service.Service, True)

    def test_service_find(self):
        self.verify_find(self.proxy.find_service, service.Service)

    def test_service_get(self):
        self.verify_get(self.proxy.get_service, service.Service)

    def test_services(self):
        self.verify_list(self.proxy.services, service.Service)

    def test_service_update(self):
        self.verify_update(self.proxy.update_service, service.Service)


class TestIdentityProxyUser(TestIdentityProxyBase):

    def test_user_create_attrs(self):
        self.verify_create(self.proxy.create_user, user.User)

    def test_user_delete(self):
        self.verify_delete(self.proxy.delete_user, user.User, False)

    def test_user_delete_ignore(self):
        self.verify_delete(self.proxy.delete_user, user.User, True)

    def test_user_find(self):
        self.verify_find(self.proxy.find_user, user.User)

    def test_user_get(self):
        self.verify_get(self.proxy.get_user, user.User)

    def test_users(self):
        self.verify_list(self.proxy.users, user.User)

    def test_user_update(self):
        self.verify_update(self.proxy.update_user, user.User)


class TestIdentityProxyTrust(TestIdentityProxyBase):

    def test_trust_create_attrs(self):
        self.verify_create(self.proxy.create_trust, trust.Trust)

    def test_trust_delete(self):
        self.verify_delete(self.proxy.delete_trust, trust.Trust, False)

    def test_trust_delete_ignore(self):
        self.verify_delete(self.proxy.delete_trust, trust.Trust, True)

    def test_trust_find(self):
        self.verify_find(self.proxy.find_trust, trust.Trust)

    def test_trust_get(self):
        self.verify_get(self.proxy.get_trust, trust.Trust)

    def test_trusts(self):
        self.verify_list(self.proxy.trusts, trust.Trust)


class TestIdentityProxyRegion(TestIdentityProxyBase):

    def test_region_create_attrs(self):
        self.verify_create(self.proxy.create_region, region.Region)

    def test_region_delete(self):
        self.verify_delete(self.proxy.delete_region, region.Region, False)

    def test_region_delete_ignore(self):
        self.verify_delete(self.proxy.delete_region, region.Region, True)

    def test_region_find(self):
        self.verify_find(self.proxy.find_region, region.Region)

    def test_region_get(self):
        self.verify_get(self.proxy.get_region, region.Region)

    def test_regions(self):
        self.verify_list(self.proxy.regions, region.Region)

    def test_region_update(self):
        self.verify_update(self.proxy.update_region, region.Region)


class TestIdentityProxyRole(TestIdentityProxyBase):

    def test_role_create_attrs(self):
        self.verify_create(self.proxy.create_role, role.Role)

    def test_role_delete(self):
        self.verify_delete(self.proxy.delete_role, role.Role, False)

    def test_role_delete_ignore(self):
        self.verify_delete(self.proxy.delete_role, role.Role, True)

    def test_role_find(self):
        self.verify_find(self.proxy.find_role, role.Role)

    def test_role_get(self):
        self.verify_get(self.proxy.get_role, role.Role)

    def test_roles(self):
        self.verify_list(self.proxy.roles, role.Role)

    def test_role_update(self):
        self.verify_update(self.proxy.update_role, role.Role)


class TestIdentityProxyRoleAssignments(TestIdentityProxyBase):

    def test_assign_domain_role_to_user(self):
        self._verify(
            "openstack.identity.v3.domain.Domain.assign_role_to_user",
            self.proxy.assign_domain_role_to_user,
            method_args=['dom_id'],
            method_kwargs={'user': 'uid', 'role': 'rid'},
            expected_args=[
                self.proxy,
                self.proxy._get_resource(user.User, 'uid'),
                self.proxy._get_resource(role.Role, 'rid')
            ]
        )

    def test_unassign_domain_role_from_user(self):
        self._verify(
            "openstack.identity.v3.domain.Domain.unassign_role_from_user",
            self.proxy.unassign_domain_role_from_user,
            method_args=['dom_id'],
            method_kwargs={'user': 'uid', 'role': 'rid'},
            expected_args=[
                self.proxy,
                self.proxy._get_resource(user.User, 'uid'),
                self.proxy._get_resource(role.Role, 'rid')
            ]
        )

    def test_validate_user_has_domain_role(self):
        self._verify(
            "openstack.identity.v3.domain.Domain.validate_user_has_role",
            self.proxy.validate_user_has_domain_role,
            method_args=['dom_id'],
            method_kwargs={'user': 'uid', 'role': 'rid'},
            expected_args=[
                self.proxy,
                self.proxy._get_resource(user.User, 'uid'),
                self.proxy._get_resource(role.Role, 'rid')
            ]
        )

    def test_assign_domain_role_to_group(self):
        self._verify(
            "openstack.identity.v3.domain.Domain.assign_role_to_group",
            self.proxy.assign_domain_role_to_group,
            method_args=['dom_id'],
            method_kwargs={'group': 'uid', 'role': 'rid'},
            expected_args=[
                self.proxy,
                self.proxy._get_resource(group.Group, 'uid'),
                self.proxy._get_resource(role.Role, 'rid')
            ]
        )

    def test_unassign_domain_role_from_group(self):
        self._verify(
            "openstack.identity.v3.domain.Domain.unassign_role_from_group",
            self.proxy.unassign_domain_role_from_group,
            method_args=['dom_id'],
            method_kwargs={'group': 'uid', 'role': 'rid'},
            expected_args=[
                self.proxy,
                self.proxy._get_resource(group.Group, 'uid'),
                self.proxy._get_resource(role.Role, 'rid')
            ]
        )

    def test_validate_group_has_domain_role(self):
        self._verify(
            "openstack.identity.v3.domain.Domain.validate_group_has_role",
            self.proxy.validate_group_has_domain_role,
            method_args=['dom_id'],
            method_kwargs={'group': 'uid', 'role': 'rid'},
            expected_args=[
                self.proxy,
                self.proxy._get_resource(group.Group, 'uid'),
                self.proxy._get_resource(role.Role, 'rid')
            ]
        )

    def test_assign_project_role_to_user(self):
        self._verify(
            "openstack.identity.v3.project.Project.assign_role_to_user",
            self.proxy.assign_project_role_to_user,
            method_args=['dom_id'],
            method_kwargs={'user': 'uid', 'role': 'rid'},
            expected_args=[
                self.proxy,
                self.proxy._get_resource(user.User, 'uid'),
                self.proxy._get_resource(role.Role, 'rid')
            ]
        )

    def test_unassign_project_role_from_user(self):
        self._verify(
            "openstack.identity.v3.project.Project.unassign_role_from_user",
            self.proxy.unassign_project_role_from_user,
            method_args=['dom_id'],
            method_kwargs={'user': 'uid', 'role': 'rid'},
            expected_args=[
                self.proxy,
                self.proxy._get_resource(user.User, 'uid'),
                self.proxy._get_resource(role.Role, 'rid')
            ]
        )

    def test_validate_user_has_project_role(self):
        self._verify(
            "openstack.identity.v3.project.Project.validate_user_has_role",
            self.proxy.validate_user_has_project_role,
            method_args=['dom_id'],
            method_kwargs={'user': 'uid', 'role': 'rid'},
            expected_args=[
                self.proxy,
                self.proxy._get_resource(user.User, 'uid'),
                self.proxy._get_resource(role.Role, 'rid')
            ]
        )

    def test_assign_project_role_to_group(self):
        self._verify(
            "openstack.identity.v3.project.Project.assign_role_to_group",
            self.proxy.assign_project_role_to_group,
            method_args=['dom_id'],
            method_kwargs={'group': 'uid', 'role': 'rid'},
            expected_args=[
                self.proxy,
                self.proxy._get_resource(group.Group, 'uid'),
                self.proxy._get_resource(role.Role, 'rid')
            ]
        )

    def test_unassign_project_role_from_group(self):
        self._verify(
            "openstack.identity.v3.project.Project.unassign_role_from_group",
            self.proxy.unassign_project_role_from_group,
            method_args=['dom_id'],
            method_kwargs={'group': 'uid', 'role': 'rid'},
            expected_args=[
                self.proxy,
                self.proxy._get_resource(group.Group, 'uid'),
                self.proxy._get_resource(role.Role, 'rid')
            ]
        )

    def test_validate_group_has_project_role(self):
        self._verify(
            "openstack.identity.v3.project.Project.validate_group_has_role",
            self.proxy.validate_group_has_project_role,
            method_args=['dom_id'],
            method_kwargs={'group': 'uid', 'role': 'rid'},
            expected_args=[
                self.proxy,
                self.proxy._get_resource(group.Group, 'uid'),
                self.proxy._get_resource(role.Role, 'rid')
            ]
        )

    def test_assign_system_role_to_user(self):
        self._verify(
            "openstack.identity.v3.system.System.assign_role_to_user",
            self.proxy.assign_system_role_to_user,
            method_kwargs={'user': 'uid', 'role': 'rid', 'system': 'all'},
            expected_args=[
                self.proxy,
                self.proxy._get_resource(user.User, 'uid'),
                self.proxy._get_resource(role.Role, 'rid')
            ]
        )

    def test_unassign_system_role_from_user(self):
        self._verify(
            "openstack.identity.v3.system.System.unassign_role_from_user",
            self.proxy.unassign_system_role_from_user,
            method_kwargs={'user': 'uid', 'role': 'rid', 'system': 'all'},
            expected_args=[
                self.proxy,
                self.proxy._get_resource(user.User, 'uid'),
                self.proxy._get_resource(role.Role, 'rid')
            ]
        )

    def test_validate_user_has_system_role(self):
        self._verify(
            "openstack.identity.v3.system.System.validate_user_has_role",
            self.proxy.validate_user_has_system_role,
            method_kwargs={'user': 'uid', 'role': 'rid', 'system': 'all'},
            expected_args=[
                self.proxy,
                self.proxy._get_resource(user.User, 'uid'),
                self.proxy._get_resource(role.Role, 'rid')
            ]
        )

    def test_assign_system_role_to_group(self):
        self._verify(
            "openstack.identity.v3.system.System.assign_role_to_group",
            self.proxy.assign_system_role_to_group,
            method_kwargs={'group': 'uid', 'role': 'rid', 'system': 'all'},
            expected_args=[
                self.proxy,
                self.proxy._get_resource(group.Group, 'uid'),
                self.proxy._get_resource(role.Role, 'rid')
            ]
        )

    def test_unassign_system_role_from_group(self):
        self._verify(
            "openstack.identity.v3.system.System.unassign_role_from_group",
            self.proxy.unassign_system_role_from_group,
            method_kwargs={'group': 'uid', 'role': 'rid', 'system': 'all'},
            expected_args=[
                self.proxy,
                self.proxy._get_resource(group.Group, 'uid'),
                self.proxy._get_resource(role.Role, 'rid')
            ]
        )

    def test_validate_group_has_system_role(self):
        self._verify(
            "openstack.identity.v3.system.System.validate_group_has_role",
            self.proxy.validate_group_has_system_role,
            method_kwargs={'group': 'uid', 'role': 'rid', 'system': 'all'},
            expected_args=[
                self.proxy,
                self.proxy._get_resource(group.Group, 'uid'),
                self.proxy._get_resource(role.Role, 'rid')
            ]
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.network.v2 import qos_rule_type
from openstack.tests.unit import base


EXAMPLE = {
    'type': 'bandwidth_limit',
    'drivers': [{
        'name': 'openvswitch',
        'supported_parameters': [{
            'parameter_values': {'start': 0, 'end': 2147483647},
            'parameter_type': 'range',
            'parameter_name': 'max_kbps'
        }, {
            'parameter_values': ['ingress', 'egress'],
            'parameter_type': 'choices',
            'parameter_name': 'direction'
        }, {
            'parameter_values': {'start': 0, 'end': 2147483647},
            'parameter_type': 'range',
            'parameter_name': 'max_burst_kbps'
        }]
    }]
}


class TestQoSRuleType(base.TestCase):

    def test_basic(self):
        sot = qos_rule_type.QoSRuleType()
        self.assertEqual('rule_type', sot.resource_key)
        self.assertEqual('rule_types', sot.resources_key)
        self.assertEqual('/qos/rule-types', sot.base_path)
        self.assertFalse(sot.allow_create)
        self.assertTrue(sot.allow_fetch)
        self.assertFalse(sot.allow_commit)
        self.assertFalse(sot.allow_delete)
        self.assertTrue(sot.allow_list)
        self.assertEqual({'type': 'type',
                          'drivers': 'drivers',
                          'all_rules': 'all_rules',
                          'all_supported': 'all_supported',
                          'limit': 'limit',
                          'marker': 'marker',
                          },
                         sot._query_mapping._mapping)

    def test_make_it(self):
        sot = qos_rule_type.QoSRuleType(**EXAMPLE)
        self.assertEqual(EXAMPLE['type'], sot.type)
        self.assertEqual(EXAMPLE['drivers'], sot.drivers)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.accelerator.v2 import _proxy
from openstack.accelerator.v2 import accelerator_request
from openstack.accelerator.v2 import deployable
from openstack.accelerator.v2 import device_profile
from openstack.tests.unit import test_proxy_base as test_proxy_base


class TestAcceleratorProxy(test_proxy_base.TestProxyBase):
    def setUp(self):
        super(TestAcceleratorProxy, self).setUp()
        self.proxy = _proxy.Proxy(self.session)


class TestAcceleratorDeployable(TestAcceleratorProxy):
    def test_list_deployables(self):
        self.verify_list(self.proxy.deployables, deployable.Deployable)


class TestAcceleratorDevice(TestAcceleratorProxy):
    def test_list_device_profile(self):
        self.verify_list(self.proxy.device_profiles,
                         device_profile.DeviceProfile)

    def test_create_device_profile(self):
        self.verify_create(self.proxy.create_device_profile,
                           device_profile.DeviceProfile)

    def test_delete_device_profile(self):
        self.verify_delete(self.proxy.delete_device_profile,
                           device_profile.DeviceProfile, False)

    def test_delete_device_profile_ignore(self):
        self.verify_delete(self.proxy.delete_device_profile,
                           device_profile.DeviceProfile, True)

    def test_get_device_profile(self):
        self.verify_get(self.proxy.get_device_profile,
                        device_profile.DeviceProfile)


class TestAcceleratorRequest(TestAcceleratorProxy):
    def test_list_accelerator_request(self):
        self.verify_list(self.proxy.accelerator_requests,
                         accelerator_request.AcceleratorRequest)

    def test_create_accelerator_request(self):
        self.verify_create(self.proxy.create_accelerator_request,
                           accelerator_request.AcceleratorRequest)

    def test_delete_accelerator_request(self):
        self.verify_delete(self.proxy.delete_accelerator_request,
                           accelerator_request.AcceleratorRequest, False)

    def test_delete_accelerator_request_ignore(self):
        self.verify_delete(self.proxy.delete_accelerator_request,
                           accelerator_request.AcceleratorRequest, True)

    def test_get_accelerator_request(self):
        self.verify_get(self.proxy.get_accelerator_request,
                        accelerator_request.AcceleratorRequest)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack import resource


class ServerActionEvent(resource.Resource):

    # Added the 'details' field in 2.84
    _max_microversion = '2.84'

    #: The name of the event
    event = resource.Body('event')
    #: The date and time when the event was started. The date and time stamp
    #: format is ISO 8601
    start_time = resource.Body('start_time')
    #: The date and time when the event finished. The date and time stamp
    #: format is ISO 8601
    finish_time = resource.Body('finish_time')
    #: The result of the event
    result = resource.Body('result')
    #: The traceback stack if an error occurred in this event.
    #: This is only visible to cloud admins by default.
    traceback = resource.Body('traceback')
    #: The name of the host on which the event occurred.
    #: This is only visible to cloud admins by default.
    host = resource.Body('host')
    #: An obfuscated hashed host ID string, or the empty string if there is no
    #: host for the event. This is a hashed value so will not actually look
    #: like a hostname, and is hashed with data from the project_id, so the
    #: same physical host as seen by two different project_ids will be
    #: different. This is useful when within the same project you need to
    #: determine if two events occurred on the same or different physical
    #: hosts.
    host_id = resource.Body('hostId')
    #: Details of the event. May be unset.
    details = resource.Body('details')


class ServerAction(resource.Resource):
    resource_key = 'instanceAction'
    resources_key = 'instanceActions'
    base_path = '/servers/{server_id}/os-instance-actions'

    # capabilities
    allow_fetch = True
    allow_list = True

    # Properties

    #: The ID of the server that this action relates to.
    server_id = resource.URI('server_id')

    #: The name of the action.
    action = resource.Body('action')
    # FIXME(stephenfin): This conflicts since there is a server ID in the URI
    # *and* in the body. We need a field that handles both or we need to use
    # different names.
    # #: The ID of the server that this action relates to.
    # server_id = resource.Body('instance_uuid')
    #: The ID of the request that this action related to.
    request_id = resource.Body('request_id')
    #: The ID of the user which initiated the server action.
    user_id = resource.Body('user_id')
    #: The ID of the project that this server belongs to.
    project_id = resource.Body('project_id')
    start_time = resource.Body('start_time')
    #: The related error message for when an action fails.
    message = resource.Body('message')
    #: Events
    events = resource.Body('events', type=list, list_type=ServerActionEvent)

    # events.details field added in 2.84
    _max_microversion = '2.84'

    _query_mapping = resource.QueryParameters(
        changes_since="changes-since",
        changes_before="changes-before",
    )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack import proxy
from openstack import resource
from openstack.shared_file_system.v2 import (
    availability_zone as _availability_zone)
from openstack.shared_file_system.v2 import (
    share_snapshot as _share_snapshot
)
from openstack.shared_file_system.v2 import (
    storage_pool as _storage_pool
)
from openstack.shared_file_system.v2 import (
    user_message as _user_message
)
from openstack.shared_file_system.v2 import limit as _limit
from openstack.shared_file_system.v2 import share as _share


class Proxy(proxy.Proxy):

    def availability_zones(self):
        """Retrieve shared file system availability zones

        :returns: A generator of availability zone resources
        :rtype:
            :class:`~openstack.shared_file_system.v2.availability_zone.AvailabilityZone`
        """
        return self._list(_availability_zone.AvailabilityZone)

    def shares(self, details=True, **query):
        """Lists all shares with details

        :param kwargs query: Optional query parameters to be sent to limit
            the shares being returned.  Available parameters include:

            * status: Filters by a share status
            * share_server_id: The UUID of the share server.
            * metadata: One or more metadata key and value pairs as a url
              encoded dictionary of strings.
            * extra_specs: The extra specifications as a set of one or more
              key-value pairs.
            * share_type_id: The UUID of a share type to query resources by.
            * name: The user defined name of the resource to filter resources
              by.
            * snapshot_id: The UUID of the share’s base snapshot to filter
              the request based on.
            * host: The host name of the resource to query with.
            * share_network_id: The UUID of the share network to filter
              resources by.
            * project_id: The ID of the project that owns the resource.
            * is_public: A boolean query parameter that, when set to true,
              allows retrieving public resources that belong to
              all projects.
            * share_group_id: The UUID of a share group to filter resource.
            * export_location_id: The export location UUID that can be used
              to filter shares or share instances.
            * export_location_path: The export location path that can be used
              to filter shares or share instances.
            * name~: The name pattern that can be used to filter shares, share
              snapshots, share networks or share groups.
            * description~: The description pattern that can be used to filter
              shares, share snapshots, share networks or share groups.
            * with_count: Whether to show count in API response or not,
              default is False.
            * limit: The maximum number of shares to return.
            * offset: The offset to define start point of share or share group
              listing.
            * sort_key: The key to sort a list of shares.
            * sort_dir: The direction to sort a list of shares. A valid value
              is asc, or desc.

        :returns: Details of shares resources
        :rtype: :class:`~openstack.shared_file_system.v2.share.Share`
        """
        base_path = '/shares/detail' if details else None
        return self._list(_share.Share, base_path=base_path, **query)

    def get_share(self, share_id):
        """Lists details of a single share

        :param share: The ID of the share to get
        :returns: Details of the identified share
        :rtype: :class:`~openstack.shared_file_system.v2.share.Share`
        """
        return self._get(_share.Share, share_id)

    def delete_share(self, share, ignore_missing=True):
        """Deletes a single share

        :param share: The ID of the share to delete
        :returns: Result of the ``delete``
        :rtype: ``None``
        """
        self._delete(_share.Share, share,
                     ignore_missing=ignore_missing)

    def update_share(self, share_id, **attrs):
        """Updates details of a single share.

        :param share: The ID of the share to update
        :param dict attrs: The attributes to update on the share
        :returns: the updated share
        :rtype: :class:`~openstack.shared_file_system.v2.share.Share`
        """
        return self._update(_share.Share, share_id, **attrs)

    def create_share(self, **attrs):
        """Creates a share from attributes

        :returns: Details of the new share
        :param dict attrs: Attributes which will be used to create
            a :class:`~openstack.shared_file_system.v2.shares.Shares`,
            comprised of the properties on the Shares class. 'size' and 'share'
            are required to create a share.
        :rtype: :class:`~openstack.shared_file_system.v2.share.Share`
        """
        return self._create(_share.Share, **attrs)

    def revert_share_to_snapshot(self, share_id, snapshot_id):
        """Reverts a share to the specified snapshot, which must be
            the most recent one known to manila.

        :param share_id: The ID of the share to revert
        :param snapshot_id: The ID of the snapshot to revert to
        :returns: Result of the ``revert``
        :rtype: ``None``
        """
        res = self._get(_share.Share, share_id)
        res.revert_to_snapshot(self, snapshot_id)

    def wait_for_status(self, res, status='active', failures=None,
                        interval=2, wait=120):
        """Wait for a resource to be in a particular status.
        :param res: The resource to wait on to reach the specified status.
            The resource must have a ``status`` attribute.
        :type resource: A :class:`~openstack.resource.Resource` object.
        :param status: Desired status.
        :param failures: Statuses that would be interpreted as failures.
        :type failures: :py:class:`list`
        :param interval: Number of seconds to wait before to consecutive
            checks. Default to 2.
        :param wait: Maximum number of seconds to wait before the change.
            Default to 120.
        :returns: The resource is returned on success.
        :raises: :class:`~openstack.exceptions.ResourceTimeout` if transition
            to the desired status failed to occur in specified seconds.
        :raises: :class:`~openstack.exceptions.ResourceFailure` if the resource
            has transited to one of the failure statuses.
        :raises: :class:`~AttributeError` if the resource does not have a
            ``status`` attribute.
        """
        failures = [] if failures is None else failures
        return resource.wait_for_status(
            self, res, status, failures, interval, wait)

    def storage_pools(self, details=True, **query):
        """Lists all back-end storage pools with details

        :param kwargs query: Optional query parameters to be sent to limit
            the storage pools being returned.  Available parameters include:

            * pool_name: The pool name for the back end.
            * host_name: The host name for the back end.
            * backend_name: The name of the back end.
            * capabilities: The capabilities for the storage back end.
            * share_type: The share type name or UUID.
        :returns: A generator of manila storage pool resources
        :rtype:
            :class:`~openstack.shared_file_system.v2.storage_pool.StoragePool`
        """
        base_path = '/scheduler-stats/pools/detail' if details else None
        return self._list(
            _storage_pool.StoragePool, base_path=base_path, **query)

    def user_messages(self, **query):
        """List shared file system user messages

        :param kwargs query: Optional query parameters to be sent to limit
            the messages being returned.  Available parameters include:

            * action_id: The ID of the action during which the message
              was created.
            * detail_id: The ID of the message detail.
            * limit: The maximum number of shares to return.
            * message_level: The message level.
            * offset: The offset to define start point of share or share
              group listing.
            * sort_key: The key to sort a list of messages.
            * sort_dir: The direction to sort a list of shares.
            * project_id: The ID of the project for which the message
              was created.
            * request_id: The ID of the request during which the message
              was created.
            * resource_id: The UUID of the resource for which the message
              was created.
            * resource_type: The type of the resource for which the message
              was created.

        :returns: A generator of user message resources
        :rtype:
            :class:`~openstack.shared_file_system.v2.user_message.UserMessage`
        """
        return self._list(
            _user_message.UserMessage, **query)

    def get_user_message(self, message_id):
        """List details of a single user message

        :param message_id: The ID of the user message
        :returns: Details of the identified user message
        :rtype:
            :class:`~openstack.shared_file_system.v2.user_message.UserMessage`
        """
        return self._get(_user_message.UserMessage, message_id)

    def delete_user_message(self, message_id, ignore_missing=True):
        """Deletes a single user message

        :param message_id: The ID of the user message
        :returns: Result of the "delete" on the user message
        :rtype:
            :class:`~openstack.shared_file_system.v2.user_message.UserMessage`
        """
        return self._delete(
            _user_message.UserMessage, message_id,
            ignore_missing=ignore_missing)

    def limits(self, **query):
        """Lists all share limits.

        :param kwargs query: Optional query parameters to be sent to limit
            the share limits being returned.

        :returns: A generator of manila share limits resources
        :rtype: :class:`~openstack.shared_file_system.v2.limit.Limit`
        """
        return self._list(
            _limit.Limit, **query)

    def share_snapshots(self, details=True, **query):
        """Lists all share snapshots with details.

        :param kwargs query: Optional query parameters to be sent to limit
            the snapshots being returned.  Available parameters include:

            * project_id: The ID of the user or service making the API request.

        :returns: A generator of manila share snapshot resources
        :rtype: :class:`~openstack.shared_file_system.v2.
            share_snapshot.ShareSnapshot`
        """
        base_path = '/snapshots/detail' if details else None
        return self._list(
            _share_snapshot.ShareSnapshot, base_path=base_path, **query)

    def get_share_snapshot(self, snapshot_id):
        """Lists details of a single share snapshot

        :param snapshot_id: The ID of the snapshot to get
        :returns: Details of the identified share snapshot
        :rtype: :class:`~openstack.shared_file_system.v2.
            share_snapshot.ShareSnapshot`
        """
        return self._get(_share_snapshot.ShareSnapshot, snapshot_id)

    def create_share_snapshot(self, **attrs):
        """Creates a share snapshot from attributes

        :returns: Details of the new share snapshot
        :rtype: :class:`~openstack.shared_file_system.v2.
            share_snapshot.ShareSnapshot`
        """
        return self._create(_share_snapshot.ShareSnapshot, **attrs)

    def update_share_snapshot(self, snapshot_id, **attrs):
        """Updates details of a single share.

        :param snapshot_id: The ID of the snapshot to update
        :pram dict attrs: The attributes to update on the snapshot
        :returns: the updated share snapshot
        :rtype: :class:`~openstack.shared_file_system.v2.
            share_snapshot.ShareSnapshot`
        """
        return self._update(_share_snapshot.ShareSnapshot, snapshot_id,
                            **attrs)

    def delete_share_snapshot(self, snapshot_id, ignore_missing=True):
        """Deletes a single share snapshot

        :param snapshot_id: The ID of the snapshot to delete
        :returns: Result of the ``delete``
        :rtype: ``None``
        """
        self._delete(_share_snapshot.ShareSnapshot, snapshot_id,
                     ignore_missing=ignore_missing)

    def wait_for_delete(self, res, interval=2, wait=120):
        """Wait for a resource to be deleted.
        :param res: The resource to wait on to be deleted.
        :type resource: A :class:`~openstack.resource.Resource` object.
        :param interval: Number of seconds to wait before to consecutive
            checks. Default to 2.
        :param wait: Maximum number of seconds to wait before the change.
            Default to 120.
        :returns: The resource is returned on success.
        :raises: :class:`~openstack.exceptions.ResourceTimeout` if transition
                 to delete failed to occur in the specified seconds.
        """
        return resource.wait_for_delete(self, res, interval, wait)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy
import uuid

from openstack.tests.unit import base


DEP_UUID = uuid.uuid4().hex
DEP_DICT = {
    'uuid': DEP_UUID,
    'name': 'dep_name',
    'parent_id': None,
    'root_id': 1,
    'num_accelerators': 4,
    'device_id': 0
}

DEV_UUID = uuid.uuid4().hex
DEV_DICT = {
    'id': 1,
    'uuid': DEV_UUID,
    'name': 'dev_name',
    'type': 'test_type',
    'vendor': '0x8086',
    'model': 'test_model',
    'std_board_info': '{"product_id": "0x09c4"}',
    'vendor_board_info': 'test_vb_info',
}

DEV_PROF_UUID = uuid.uuid4().hex
DEV_PROF_GROUPS = [
    {"resources:ACCELERATOR_FPGA": "1",
     "trait:CUSTOM_FPGA_INTEL_PAC_ARRIA10": "required",
     "trait:CUSTOM_FUNCTION_ID_3AFB": "required",
     },
    {"resources:CUSTOM_ACCELERATOR_FOO": "2",
     "resources:CUSTOM_MEMORY": "200",
     "trait:CUSTOM_TRAIT_ALWAYS": "required",
     }
]
DEV_PROF_DICT = {
    "id": 1,
    "uuid": DEV_PROF_UUID,
    "name": 'afaas_example_1',
    "groups": DEV_PROF_GROUPS,
}

NEW_DEV_PROF_DICT = copy.copy(DEV_PROF_DICT)

ARQ_UUID = uuid.uuid4().hex
ARQ_DEV_RP_UUID = uuid.uuid4().hex
ARQ_INSTANCE_UUID = uuid.uuid4().hex
ARQ_ATTACH_INFO_STR = '{"bus": "5e", '\
    '"device": "00", '\
    '"domain": "0000", '\
    '"function": "1"}'
ARQ_DICT = {
    'uuid': ARQ_UUID,
    'hostname': 'test_hostname',
    'device_profile_name': 'fake-devprof',
    'device_profile_group_id': 0,
    'device_rp_uuid': ARQ_DEV_RP_UUID,
    'instance_uuid': ARQ_INSTANCE_UUID,
    'attach_handle_type': 'PCI',
    'attach_handle_info': ARQ_ATTACH_INFO_STR,
}

NEW_ARQ_DICT = copy.copy(ARQ_DICT)


class TestAccelerator(base.TestCase):
    def setUp(self):
        super(TestAccelerator, self).setUp()
        self.use_cyborg()

    def test_list_deployables(self):
        self.register_uris([
            dict(method='GET',
                 uri=self.get_mock_url(
                     'accelerator',
                     'public',
                     append=['v2', 'deployables']),
                 json={'deployables': [DEP_DICT]}
                 ),
        ])
        dep_list = self.cloud.list_deployables()
        self.assertEqual(len(dep_list), 1)
        self.assertEqual(dep_list[0].id, DEP_DICT['uuid'])
        self.assertEqual(dep_list[0].name, DEP_DICT['name'])
        self.assertEqual(dep_list[0].parent_id, DEP_DICT['parent_id'])
        self.assertEqual(dep_list[0].root_id, DEP_DICT['root_id'])
        self.assertEqual(dep_list[0].num_accelerators,
                         DEP_DICT['num_accelerators'])
        self.assertEqual(dep_list[0].device_id, DEP_DICT['device_id'])
        self.assert_calls()

    def test_list_devices(self):
        self.register_uris([
            dict(method='GET',
                 uri=self.get_mock_url(
                     'accelerator',
                     'public',
                     append=['v2', 'devices']),
                 json={'devices': [DEV_DICT]}
                 ),
        ])
        dev_list = self.cloud.list_devices()
        self.assertEqual(len(dev_list), 1)
        self.assertEqual(dev_list[0].id, DEV_DICT['id'])
        self.assertEqual(dev_list[0].uuid, DEV_DICT['uuid'])
        self.assertEqual(dev_list[0].name, DEV_DICT['name'])
        self.assertEqual(dev_list[0].type, DEV_DICT['type'])
        self.assertEqual(dev_list[0].vendor, DEV_DICT['vendor'])
        self.assertEqual(dev_list[0].model, DEV_DICT['model'])
        self.assertEqual(dev_list[0].std_board_info,
                         DEV_DICT['std_board_info'])
        self.assertEqual(dev_list[0].vendor_board_info,
                         DEV_DICT['vendor_board_info'])
        self.assert_calls()

    def test_list_device_profiles(self):
        self.register_uris([
            dict(method='GET',
                 uri=self.get_mock_url(
                     'accelerator',
                     'public',
                     append=['v2', 'device_profiles']),
                 json={'device_profiles': [DEV_PROF_DICT]}
                 ),
        ])
        dev_prof_list = self.cloud.list_device_profiles()
        self.assertEqual(len(dev_prof_list), 1)
        self.assertEqual(dev_prof_list[0].id, DEV_PROF_DICT['id'])
        self.assertEqual(dev_prof_list[0].uuid, DEV_PROF_DICT['uuid'])
        self.assertEqual(dev_prof_list[0].name, DEV_PROF_DICT['name'])
        self.assertEqual(dev_prof_list[0].groups, DEV_PROF_DICT['groups'])
        self.assert_calls()

    def test_create_device_profile(self):
        self.register_uris([
            dict(method='POST',
                 uri=self.get_mock_url(
                     'accelerator',
                     'public',
                     append=['v2', 'device_profiles']),
                 json=NEW_DEV_PROF_DICT)
        ])

        attrs = {
            'name': NEW_DEV_PROF_DICT['name'],
            'groups': NEW_DEV_PROF_DICT['groups']
        }

        self.assertTrue(
            self.cloud.create_device_profile(
                attrs
            )
        )
        self.assert_calls()

    def test_delete_device_profile(self, filters=None):
        self.register_uris([
            dict(method='GET',
                 uri=self.get_mock_url(
                     'accelerator',
                     'public',
                     append=['v2', 'device_profiles', DEV_PROF_DICT['name']]),
                 json={"device_profiles": [DEV_PROF_DICT]}),
            dict(method='DELETE',
                 uri=self.get_mock_url(
                     'accelerator',
                     'public',
                     append=['v2', 'device_profiles', DEV_PROF_DICT['name']]),
                 json=DEV_PROF_DICT)

        ])
        self.assertTrue(
            self.cloud.delete_device_profile(
                DEV_PROF_DICT['name'],
                filters
            )
        )
        self.assert_calls()

    def test_list_accelerator_requests(self):
        self.register_uris([
            dict(method='GET',
                 uri=self.get_mock_url(
                     'accelerator',
                     'public',
                     append=['v2', 'accelerator_requests']),
                 json={'arqs': [ARQ_DICT]}
                 ),
        ])
        arq_list = self.cloud.list_accelerator_requests()
        self.assertEqual(len(arq_list), 1)
        self.assertEqual(arq_list[0].uuid, ARQ_DICT['uuid'])
        self.assertEqual(arq_list[0].device_profile_name,
                         ARQ_DICT['device_profile_name'])
        self.assertEqual(arq_list[0].device_profile_group_id,
                         ARQ_DICT['device_profile_group_id'])
        self.assertEqual(arq_list[0].device_rp_uuid,
                         ARQ_DICT['device_rp_uuid'])
        self.assertEqual(arq_list[0].instance_uuid,
                         ARQ_DICT['instance_uuid'])
        self.assertEqual(arq_list[0].attach_handle_type,
                         ARQ_DICT['attach_handle_type'])
        self.assertEqual(arq_list[0].attach_handle_info,
                         ARQ_DICT['attach_handle_info'])
        self.assert_calls()

    def test_create_accelerator_request(self):
        self.register_uris([
            dict(method='POST',
                 uri=self.get_mock_url(
                     'accelerator',
                     'public',
                     append=['v2', 'accelerator_requests']),
                 json=NEW_ARQ_DICT
                 ),
        ])

        attrs = {
            'device_profile_name': NEW_ARQ_DICT['device_profile_name'],
            'device_profile_group_id': NEW_ARQ_DICT['device_profile_group_id']
        }

        self.assertTrue(
            self.cloud.create_accelerator_request(
                attrs
            )
        )
        self.assert_calls()

    def test_delete_accelerator_request(self, filters=None):
        self.register_uris([
            dict(method='GET',
                 uri=self.get_mock_url(
                     'accelerator',
                     'public',
                     append=['v2', 'accelerator_requests', ARQ_DICT['uuid']]),
                 json={"accelerator_requests": [ARQ_DICT]}),
            dict(method='DELETE',
                 uri=self.get_mock_url(
                     'accelerator',
                     'public',
                     append=['v2', 'accelerator_requests', ARQ_DICT['uuid']]),
                 json=ARQ_DICT)

        ])
        self.assertTrue(
            self.cloud.delete_accelerator_request(
                ARQ_DICT['uuid'],
                filters
            )
        )
        self.assert_calls()

    def test_bind_accelerator_request(self):
        self.register_uris([
            dict(method='GET',
                 uri=self.get_mock_url(
                     'accelerator',
                     'public',
                     append=['v2', 'accelerator_requests', ARQ_DICT['uuid']]),
                 json={"accelerator_requests": [ARQ_DICT]}),
            dict(method='PATCH',
                 uri=self.get_mock_url(
                     'accelerator',
                     'public',
                     append=['v2', 'accelerator_requests', ARQ_DICT['uuid']]),
                 json=ARQ_DICT)
        ])
        properties = [{'path': '/hostname',
                       'value': ARQ_DICT['hostname'],
                       'op': 'add'},
                      {'path': '/instance_uuid',
                       'value': ARQ_DICT['instance_uuid'],
                       'op': 'add'},
                      {'path': '/device_rp_uuid',
                       'value': ARQ_DICT['device_rp_uuid'],
                       'op': 'add'}]

        self.assertTrue(
            self.cloud.bind_accelerator_request(
                ARQ_DICT['uuid'], properties
            )
        )
        self.assert_calls()

    def test_unbind_accelerator_request(self):
        self.register_uris([
            dict(method='GET',
                 uri=self.get_mock_url(
                     'accelerator',
                     'public',
                     append=['v2', 'accelerator_requests', ARQ_DICT['uuid']]),
                 json={"accelerator_requests": [ARQ_DICT]}),
            dict(method='PATCH',
                 uri=self.get_mock_url(
                     'accelerator',
                     'public',
                     append=['v2', 'accelerator_requests', ARQ_DICT['uuid']]),
                 json=ARQ_DICT)
        ])

        properties = [{'path': '/hostname',
                       'op': 'remove'},
                      {'path': '/instance_uuid',
                       'op': 'remove'},
                      {'path': '/device_rp_uuid',
                       'op': 'remove'}]

        self.assertTrue(
            self.cloud.unbind_accelerator_request(
                ARQ_DICT['uuid'], properties
            )
        )
        self.assert_calls()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.clustering.v1 import event
from openstack.tests.unit import base


FAKE = {
    'action': 'NODE_CREATE',
    'cluster_id': None,
    'id': 'ffaed25e-46f5-4089-8e20-b3b4722fd597',
    'level': '20',
    'oid': 'efff1c11-2ada-47da-bedd-2c9af4fd099a',
    'oname': 'node_create_b4a49016',
    'otype': 'NODEACTION',
    'project': '42d9e9663331431f97b75e25136307ff',
    'status': 'START',
    'status_reason': 'The action was abandoned.',
    'timestamp': '2016-10-10T12:46:36.000000',
    'user': '5e5bf8027826429c96af157f68dc9072',
    'meta_data': {
        "action": {
            "created_at": "2019-07-13T13:18:18Z",
            "outputs": {}
        }
    }
}


class TestEvent(base.TestCase):

    def setUp(self):
        super(TestEvent, self).setUp()

    def test_basic(self):
        sot = event.Event()
        self.assertEqual('event', sot.resource_key)
        self.assertEqual('events', sot.resources_key)
        self.assertEqual('/events', sot.base_path)
        self.assertTrue(sot.allow_fetch)
        self.assertTrue(sot.allow_list)

    def test_instantiate(self):
        sot = event.Event(**FAKE)
        self.assertEqual(FAKE['id'], sot.id)
        self.assertEqual(FAKE['action'], sot.action)
        self.assertEqual(FAKE['cluster_id'], sot.cluster_id)
        self.assertEqual(FAKE['level'], sot.level)
        self.assertEqual(FAKE['oid'], sot.obj_id)
        self.assertEqual(FAKE['oname'], sot.obj_name)
        self.assertEqual(FAKE['otype'], sot.obj_type)
        self.assertEqual(FAKE['project'], sot.project_id)
        self.assertEqual(FAKE['status'], sot.status)
        self.assertEqual(FAKE['status_reason'], sot.status_reason)
        self.assertEqual(FAKE['timestamp'], sot.generated_at)
        self.assertEqual(FAKE['user'], sot.user_id)
        self.assertEqual(FAKE['meta_data'], sot.meta_data)



# Copyright 2017 OVH SAS
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from openstack.cloud import exc
from openstack.network.v2 import qos_rule_type
from openstack.tests.unit import base


class TestQosRuleType(base.TestCase):

    rule_type_name = "bandwidth_limit"

    qos_extension = {
        "updated": "2015-06-08T10:00:00-00:00",
        "name": "Quality of Service",
        "links": [],
        "alias": "qos",
        "description": "The Quality of Service extension."
    }
    qos_rule_type_details_extension = {
        "updated": "2017-06-22T10:00:00-00:00",
        "name": "Details of QoS rule types",
        "links": [],
        "alias": "qos-rule-type-details",
        "description": ("Expose details about QoS rule types supported by "
                        "loaded backend drivers")
    }

    mock_rule_type_bandwidth_limit = {
        'type': 'bandwidth_limit'
    }
    mock_rule_type_dscp_marking = {
        'type': 'dscp_marking'
    }
    mock_rule_types = [
        mock_rule_type_bandwidth_limit, mock_rule_type_dscp_marking]

    mock_rule_type_details = {
        'drivers': [{
            'name': 'linuxbridge',
            'supported_parameters': [{
                'parameter_values': {'start': 0, 'end': 2147483647},
                'parameter_type': 'range',
                'parameter_name': u'max_kbps'
            }, {
                'parameter_values': ['ingress', 'egress'],
                'parameter_type': 'choices',
                'parameter_name': u'direction'
            }, {
                'parameter_values': {'start': 0, 'end': 2147483647},
                'parameter_type': 'range',
                'parameter_name': 'max_burst_kbps'
            }]
        }],
        'type': rule_type_name
    }

    def _compare_rule_types(self, exp, real):
        self.assertDictEqual(
            qos_rule_type.QoSRuleType(**exp).to_dict(computed=False),
            real.to_dict(computed=False))

    def test_list_qos_rule_types(self):
        self.register_uris([
            dict(method='GET',
                 uri=self.get_mock_url(
                     'network', 'public', append=['v2.0', 'extensions']),
                 json={'extensions': [self.qos_extension]}),
            dict(method='GET',
                 uri=self.get_mock_url(
                     'network', 'public',
                     append=['v2.0', 'qos', 'rule-types']),
                 json={'rule_types': self.mock_rule_types})
        ])
        rule_types = self.cloud.list_qos_rule_types()
        for a, b in zip(self.mock_rule_types, rule_types):
            self._compare_rule_types(a, b)
        self.assert_calls()

    def test_list_qos_rule_types_no_qos_extension(self):
        self.register_uris([
            dict(method='GET',
                 uri=self.get_mock_url(
                     'network', 'public', append=['v2.0', 'extensions']),
                 json={'extensions': []})
        ])
        self.assertRaises(exc.OpenStackCloudException,
                          self.cloud.list_qos_rule_types)
        self.assert_calls()

    def test_get_qos_rule_type_details(self):
        self.register_uris([
            dict(method='GET',
                 uri=self.get_mock_url(
                     'network', 'public', append=['v2.0', 'extensions']),
                 json={'extensions': [
                     self.qos_extension,
                     self.qos_rule_type_details_extension]}),
            dict(method='GET',
                 uri=self.get_mock_url(
                     'network', 'public', append=['v2.0', 'extensions']),
                 json={'extensions': [
                     self.qos_extension,
                     self.qos_rule_type_details_extension]}),
            dict(method='GET',
                 uri=self.get_mock_url(
                     'network', 'public',
                     append=['v2.0', 'qos', 'rule-types',
                             self.rule_type_name]),
                     json={'rule_type': self.mock_rule_type_details})
        ])

        self._compare_rule_types(
            self.mock_rule_type_details,
            self.cloud.get_qos_rule_type_details(self.rule_type_name)
        )
        self.assert_calls()

    def test_get_qos_rule_type_details_no_qos_extension(self):
        self.register_uris([
            dict(method='GET',
                 uri=self.get_mock_url(
                     'network', 'public', append=['v2.0', 'extensions']),
                 json={'extensions': []})
        ])
        self.assertRaises(
            exc.OpenStackCloudException,
            self.cloud.get_qos_rule_type_details, self.rule_type_name)
        self.assert_calls()

    def test_get_qos_rule_type_details_no_qos_details_extension(self):
        self.register_uris([
            dict(method='GET',
                 uri=self.get_mock_url(
                     'network', 'public', append=['v2.0', 'extensions']),
                 json={'extensions': [self.qos_extension]}),
            dict(method='GET',
                 uri=self.get_mock_url(
                     'network', 'public', append=['v2.0', 'extensions']),
                 json={'extensions': [self.qos_extension]})
        ])
        self.assertRaises(
            exc.OpenStackCloudException,
            self.cloud.get_qos_rule_type_details, self.rule_type_name)
        self.assert_calls()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.key_manager.v1 import _format
from openstack import resource


class Container(resource.Resource):
    resources_key = 'containers'
    base_path = '/containers'

    # capabilities
    allow_create = True
    allow_fetch = True
    allow_commit = True
    allow_delete = True
    allow_list = True

    # Properties
    #: A URI for this container
    container_ref = resource.Body('container_ref')
    #: The ID for this container
    container_id = resource.Body(
        'container_ref', alternate_id=True,
        type=_format.HREFToUUID)
    #: The timestamp when this container was created.
    created_at = resource.Body('created')
    #: The name of this container
    name = resource.Body('name')
    #: A list of references to secrets in this container
    secret_refs = resource.Body('secret_refs', type=list)
    #: The status of this container
    status = resource.Body('status')
    #: The type of this container
    type = resource.Body('type')
    #: The timestamp when this container was updated.
    updated_at = resource.Body('updated')
    #: A party interested in this container.
    consumers = resource.Body('consumers', type=list)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.compute.v2 import migration
from openstack.tests.unit import base

EXAMPLE = {
    'uuid': '42341d4b-346a-40d0-83c6-5f4f6892b650',
    'instance_uuid': '9128d044-7b61-403e-b766-7547076ff6c1',
    'user_id': '78348f0e-97ee-4d70-ad34-189692673ea2',
    'project_id': '9842f0f7-1229-4355-afe7-15ebdbb8c3d8',
    'created_at': '2016-06-23T14:42:02.000000',
    'updated_at': '2016-06-23T14:42:02.000000',
    'status': 'migrating',
    'source_compute': 'compute10',
    'source_node': 'node10',
    'dest_host': '5.6.7.8',
    'dest_compute': 'compute20',
    'dest_node': 'node20',
    'migration_type': 'resize',
    'old_instance_type_id': 5,
    'new_instance_type_id': 6,
}


class TestMigration(base.TestCase):

    def test_basic(self):
        sot = migration.Migration()
        self.assertIsNone(sot.resource_key)  # we don't support fetch
        self.assertEqual('migrations', sot.resources_key)
        self.assertEqual('/os-migrations', sot.base_path)
        self.assertFalse(sot.allow_create)
        self.assertFalse(sot.allow_fetch)
        self.assertFalse(sot.allow_commit)
        self.assertFalse(sot.allow_delete)
        self.assertTrue(sot.allow_list)

        self.assertDictEqual(
            {
                'limit': 'limit',
                'marker': 'marker',
                'host': 'host',
                'status': 'status',
                'migration_type': 'migration_type',
                'source_compute': 'source_compute',
                'user_id': 'user_id',
                'project_id': 'project_id',
                'changes_since': 'changes-since',
                'changes_before': 'changes-before',
                'server_id': 'instance_uuid',
            },
            sot._query_mapping._mapping,
        )

    def test_make_it(self):
        sot = migration.Migration(**EXAMPLE)
        self.assertEqual(EXAMPLE['uuid'], sot.id)
        self.assertEqual(EXAMPLE['instance_uuid'], sot.server_id)
        self.assertEqual(EXAMPLE['user_id'], sot.user_id)
        self.assertEqual(EXAMPLE['project_id'], sot.project_id)
        self.assertEqual(EXAMPLE['created_at'], sot.created_at)
        self.assertEqual(EXAMPLE['updated_at'], sot.updated_at)
        self.assertEqual(EXAMPLE['status'], sot.status)
        self.assertEqual(EXAMPLE['source_compute'], sot.source_compute)
        self.assertEqual(EXAMPLE['source_node'], sot.source_node)
        self.assertEqual(EXAMPLE['dest_host'], sot.dest_host)
        self.assertEqual(EXAMPLE['dest_compute'], sot.dest_compute)
        self.assertEqual(EXAMPLE['dest_node'], sot.dest_node)
        self.assertEqual(EXAMPLE['migration_type'], sot.migration_type)
        self.assertEqual(EXAMPLE['old_instance_type_id'], sot.old_flavor_id)
        self.assertEqual(EXAMPLE['new_instance_type_id'], sot.new_flavor_id)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
test_project_cleanup
----------------------------------

Functional tests for project cleanup methods.
"""
import queue

from openstack.tests.functional import base


class TestProjectCleanup(base.BaseFunctionalTest):

    _wait_for_timeout_key = 'OPENSTACKSDK_FUNC_TEST_TIMEOUT_CLEANUP'

    def setUp(self):
        super(TestProjectCleanup, self).setUp()
        self.conn = self.user_cloud_alt
        self.network_name = self.getUniqueString('network')

    def _create_network_resources(self):
        conn = self.conn
        self.net = conn.network.create_network(
            name=self.network_name,
        )
        self.subnet = conn.network.create_subnet(
            name=self.getUniqueString('subnet'),
            network_id=self.net.id,
            cidr='192.169.1.0/24',
            ip_version=4,
        )
        self.router = conn.network.create_router(
            name=self.getUniqueString('router')
        )
        conn.network.add_interface_to_router(
            self.router.id,
            subnet_id=self.subnet.id)

    def test_cleanup(self):
        self._create_network_resources()
        status_queue = queue.Queue()

        # First round - check no resources are old enough
        self.conn.project_cleanup(
            dry_run=True,
            wait_timeout=120,
            status_queue=status_queue,
            filters={'created_at': '2000-01-01'})

        self.assertTrue(status_queue.empty())

        # Second round - resource evaluation function return false, ensure
        # nothing identified
        self.conn.project_cleanup(
            dry_run=True,
            wait_timeout=120,
            status_queue=status_queue,
            filters={'created_at': '2200-01-01'},
            resource_evaluation_fn=lambda x, y, z: False)

        self.assertTrue(status_queue.empty())

        # Third round - filters set too low
        self.conn.project_cleanup(
            dry_run=True,
            wait_timeout=120,
            status_queue=status_queue,
            filters={'created_at': '2200-01-01'})

        objects = []
        while not status_queue.empty():
            objects.append(status_queue.get())

        # At least known networks should be identified
        net_names = list(obj.name for obj in objects)
        self.assertIn(self.network_name, net_names)

        # Fourth round - dry run with no filters, ensure everything identified
        self.conn.project_cleanup(
            dry_run=True,
            wait_timeout=120,
            status_queue=status_queue)

        objects = []
        while not status_queue.empty():
            objects.append(status_queue.get())

        net_names = list(obj.name for obj in objects)
        self.assertIn(self.network_name, net_names)

        # Ensure network still exists
        net = self.conn.network.get_network(self.net.id)
        self.assertEqual(net.name, self.net.name)

        # Last round - do a real cleanup
        self.conn.project_cleanup(
            dry_run=False,
            wait_timeout=600,
            status_queue=status_queue)

        objects = []
        while not status_queue.empty():
            objects.append(status_queue.get())

        nets = self.conn.network.networks()
        net_names = list(obj.name for obj in nets)
        # Since we might not have enough privs to drop all nets - ensure
        # we do not have our known one
        self.assertNotIn(self.network_name, net_names)
